# pylint: skip-file

import customtkinter as ctk

from ..evaluation import (
    calc_dipole,
    calc_dipole_moment,
    calc_exciton_transfer,
    calc_lifetime,
)

from .user_interface_utils import change_state_all_widgets

# --------------------------------------------------


class ExcitonFrame(ctk.CTkFrame):
    def __init__(self, master, **kwargs):
        """
        Notes:
            This frame is located in the plot_options tab. This means that the plot_options tab is its master.
        """

        # initialization of the ctk.CTkFrame class
        super().__init__(master)
        self.pack(fill="both", expand=True)
        self.kwargs = kwargs

        # widgets
        self.lifetime_button = ctk.CTkButton(
            self, text="Calculate Exciton Lifetime", command=self._calc_lifetime
        )
        self.lifetime_button.grid(
            row=0, column=0, columnspan=2, padx=10, pady=10, sticky="ew"
        )
        self.dipole_button = ctk.CTkButton(
            self, text="Calculate Charge Separation", command=self._calc_dipole
        )
        self.dipole_button.grid(
            row=1, column=0, columnspan=2, padx=10, pady=10, sticky="ew"
        )
        self.dipole_moment_button = ctk.CTkButton(
            self, text="Calculate Dipole Moment", command=self._calc_dipole_moment
        )
        self.dipole_moment_button.grid(
            row=2, column=0, columnspan=2, padx=10, pady=10, sticky="ew"
        )
        self.exciton_transfer_button = ctk.CTkButton(
            self,
            text="Calculate Exciton Population",
            command=self._calc_exciton_transfer,
        )
        self.exciton_transfer_button.grid(
            row=3, column=0, padx=10, pady=10, columnspan=2, sticky="ew"
        )

    def _calc_lifetime(self):
        assert (
            self.kwargs["description"] == "2P"
        ), "2P description is required for the calculation."
        assert (
            self.kwargs["relaxation"] == True
        ), "Groundstate is required for the calculation."

        lifetime = calc_lifetime(**self.kwargs)
        if isinstance(lifetime, str):
            print(f"Exciton Lifetime: {lifetime}" "\n-------------------------------")
        else:
            print(
                f"Exciton Lifetime: {lifetime} fs" "\n-------------------------------"
            )

    def _calc_dipole(self):
        assert (
            self.kwargs["description"] == "2P"
        ), "2P description is required for the calculation."
        assert (
            self.kwargs["relaxation"] == True
        ), "Groundstate is required for the calculation."

        dipole = calc_dipole(**self.kwargs)
        print(f"Charge Separation: {dipole} A" "\n-------------------------------")

    def _calc_dipole_moment(self):
        assert (
            self.kwargs["description"] == "2P"
        ), "2P description is required for the calculation."
        assert (
            self.kwargs["relaxation"] == True
        ), "Groundstate is required for the calculation."

        dipole_moment = calc_dipole_moment(**self.kwargs)
        print(f"Dipole Moment: {dipole_moment} D" "\n-------------------------------")

    def _calc_exciton_transfer(self):
        assert (
            self.kwargs["description"] == "2P"
        ), "2P description is required for the calculation."
        assert (
            self.kwargs["relaxation"] == True
        ), "Groundstate is required for the calculation."
        assert (
            "exciton" in self.kwargs["particles"]
        ), "Exciton must be selected in particles."

        avg_pop_upper_strand, avg_pop_lower_strand = calc_exciton_transfer(
            **self.kwargs
        )
        avg_pop_upper_strand, avg_pop_lower_strand = (
            avg_pop_upper_strand["exciton"],
            avg_pop_lower_strand["exciton"],
        )
        print(
            f"Average Exciton Population (upper strand): {avg_pop_upper_strand}"
            + f"\nAverage Exciton Population (lower strand): {avg_pop_lower_strand}"
            + "\n-------------------------------"
        )


# --------------------------------------------------


class PopFrame(ctk.CTkFrame):
    def __init__(self, master, tb_basis, **kwargs):
        """
        Notes:
            This frame is located in the plot_options tab. This means that the plot_options tab is its master.
        """

        # initialization of the ctk.CTkFrame class
        super().__init__(master)
        self.pack(fill="both", expand=True)

        self.explain_label = ctk.CTkLabel(
            self, text="Calculate the population of each \nbase of the DNA sequence."
        )
        self.explain_label.grid(row=0, column=0, columnspan=2, padx=10, pady=10)

        self.tb_site_label = ctk.CTkLabel(self, text="TB site:")
        self.tb_site_label.grid(row=1, column=0, padx=10, pady=10)

        self.tb_site_combo = ctk.CTkComboBox(
            self, values=["Heatmap", "All DNA Bases"] + tb_basis
        )
        self.tb_site_combo.grid(row=1, column=1, padx=10, pady=10)
        self.tb_site_combo.set("Heatmap")

    def get_pop_kwargs(self):
        """Returns the values of widgets with get() method in dictionary format."""
        pop_kwargs = {
            "init_tb_site": self.tb_site_combo.get(),
        }
        return pop_kwargs


# --------------------------------------------------


class CohFrame(ctk.CTkFrame):
    def __init__(self, master, **kwargs):
        """
        Notes:
            This frame is located in the plot_options tab. This means that the plot_options tab is its master.
        """

        # initialization of the ctk.CTkFrame class
        super().__init__(master)
        self.pack(fill="both", expand=True)

        self.explain_label = ctk.CTkLabel(
            self, text="Calculate the coherence of the DNA sequence."
        )
        self.explain_label.grid(row=0, column=0, padx=10, pady=10)

    def get_coh_kwargs(self):
        """Returns the values of widgets with get() method in dictionary format."""
        coh_kwargs = {}
        return coh_kwargs


# --------------------------------------------------


class FourierFrame(ctk.CTkFrame):
    def __init__(self, master, tb_basis, **kwargs):
        """
        Notes:
            This frame is located in the plot_options tab. This means that the plot_options tab is its master.
        """

        # initialization of the ctk.CTkFrame class
        super().__init__(master)
        self.pack(fill="both", expand=True)
        self.kwargs = kwargs

        # widgets
        self.tb_site_label = ctk.CTkLabel(self, text="TB site:")
        self.tb_site_label.grid(row=0, column=0, padx=10, pady=10)

        self.tb_site_combo = ctk.CTkComboBox(self, values=tb_basis)
        self.tb_site_combo.grid(row=0, column=1, padx=10, pady=10)

        self.init_e_site_label = ctk.CTkLabel(self, text="Initial electron \n TB site:")
        self.init_e_site_label.grid(row=1, column=0, padx=10, pady=10)

        self.init_e_site_combo = ctk.CTkComboBox(self, values=tb_basis)
        self.init_e_site_combo.grid(row=1, column=1, padx=10, pady=10)

        self.init_h_site_label = ctk.CTkLabel(self, text="Initial hole \n TB site:")
        self.init_h_site_label.grid(row=2, column=0, padx=10, pady=10)

        self.init_h_site_combo = ctk.CTkComboBox(self, values=tb_basis)
        self.init_h_site_combo.grid(row=2, column=1, padx=10, pady=10)

        self.x_axis_combo = ctk.CTkComboBox(self, values=["period", "frequency"])
        self.x_axis_combo.grid(row=3, column=0, padx=10, pady=10)
        self.x_axis_combo.set("period")

        self.average_pop_var = ctk.BooleanVar(value=False)
        self.average_pop_check = ctk.CTkCheckBox(
            self, text="Calculate average population", variable=self.average_pop_var
        )
        self.average_pop_check.grid(row=4, column=0, columnspan=2, padx=10, pady=10)

    def get_fourier_kwargs(self):
        """Returns the values of widgets with get() method in dictionary format."""

        if self.kwargs["description"] == "1P":
            if self.kwargs["particles"] == ["electron"]:
                init_state = self.init_e_site_combo.get()
            if self.kwargs["particles"] == ["hole"]:
                init_state = self.init_h_site_combo.get()
        if self.kwargs["description"] == "2P":
            init_state = (self.init_e_site_combo.get(), self.init_h_site_combo.get())

        fourier_kwargs = {
            "end_state": self.tb_site_combo.get(),
            "init_state": init_state,
            "x_axis": self.x_axis_combo.get(),
        }
        return fourier_kwargs


# --------------------------------------------------


class PlotOptionsTab(ctk.CTkTabview):
    def __init__(self, master, tb_basis, **kwargs):
        """
        Notes:
            This tab is located in the plot_options frame. This means that the plot_options frame is its master.
            The tab itself is the master for four further frames.
        """

        # initialization of the ctk.CTkTabview class
        super().__init__(master)

        self.pop_tab = self.add("Population")
        self.coh_tab = self.add("Coherence")
        self.fourier_tab = self.add("Fourier")
        self.lifetime_tab = self.add("Evaluation")
        self.set("Population")

        self.pop_frame = PopFrame(self.pop_tab, tb_basis, **kwargs)
        self.coh_frame = CohFrame(self.coh_tab, **kwargs)
        self.fourier_frame = FourierFrame(self.fourier_tab, tb_basis, **kwargs)
        self.exciton_frame = ExcitonFrame(self.lifetime_tab, **kwargs)


class PlotOptionsFrame(ctk.CTkFrame):
    def __init__(self, master, tb_basis, **kwargs):
        """
        Notes:
            This frame is located in the main window. This means that the main window is its master.
            The frame itself is the master for the tab plot_options_tab.
            The frame uses the press_second_confirm(), enable_initial_frame() from the master.
        """

        # initialization of the ctk.CTkFrame class
        super().__init__(master)

        self.label = ctk.CTkLabel(
            self, text="Plotting", font=ctk.CTkFont(size=20, weight="bold")
        )
        self.label.grid(row=0, column=0, columnspan=2, pady=10, padx=10)

        self.plot_options_tab = PlotOptionsTab(self, tb_basis, **kwargs)
        self.plot_options_tab.grid(row=1, column=0, columnspan=2, pady=10, padx=10)

        self.back_button = ctk.CTkButton(
            self, text="Back", command=master.enable_options_frame
        )
        self.back_button.grid(row=3, column=0, pady=10, padx=10)

        self.submit_button = ctk.CTkButton(self, text="Submit", command=master.submit)
        self.submit_button.grid(row=3, column=1, pady=10, padx=10)

    def change_state(self, state):
        """Changes the state of certain widgets (between 'normal' and 'disabled')."""
        change_state_all_widgets(self.plot_options_tab.pop_frame, state=state)
        change_state_all_widgets(self.plot_options_tab.coh_frame, state=state)
        change_state_all_widgets(self.plot_options_tab.fourier_frame, state=state)
        change_state_all_widgets(self.plot_options_tab.exciton_frame, state=state)
        self.submit_button.configure(state=state)
        self.back_button.configure(state=state)



import os
import json

import numpy as np

from .slater_koster import calc_orbital_overlap

# ----------------------------------------------


class Dimer:
    """
    A class to represent a dimer consisting of two molecules.

    Parameters
    ----------
    molecule1 : Molecule
        The first molecule in the dimer.
    molecule2 : Molecule
        The second molecule in the dimer.
    identifier : str, optional
        An identifier for the dimer. If not provided, it is generated from the identifiers of the two molecules.

    Attributes
    ----------
    molecule1 : Molecule
        The first molecule in the dimer.
    molecule2 : Molecule
        The second molecule in the dimer.
    identifier : str, optional
        An identifier for the dimer. If not provided, it is generated from the identifiers of the two molecules.
    H_int : numpy.ndarray
        The interaction Hamiltonian matrix between the two molecules.
    t_HOMO : float
        The hopping parameter for the highest occupied molecular orbital (HOMO).
    t_HOMO_1 : float
        The hopping parameter for the first HOMO-1.
    t_HOMO_2 : float
        The hopping parameter for the second HOMO-2.
    t_LUMO : float
        The hopping parameter for the lowest unoccupied molecular orbital (LUMO).
    t_LUMO_1 : float
        The hopping parameter for the first LUMO-1.
    t_LUMO_2 : float
        The hopping parameter for the second LUMO-2.

    Methods
    -------
    save_results(directory="results"):
        Saves the results of the calculation to a JSON file.
    """

    def __init__(self, molecule1, molecule2, identifier=None):
        self.molecule1 = molecule1
        self.molecule2 = molecule2

        # Identifier
        self.identifier = identifier
        if self.identifier is None:
            self.identifier = molecule1.identifier + molecule2.identifier

        # Interaction Hamiltonian (not the full LCAO Hamiltonian)
        self.H_int = self._calc_H_int()

        # Hopping parameters for electron and hole
        self.t_HOMO = self.molecule1.HOMO @ self.H_int @ self.molecule2.HOMO
        self.t_HOMO_1 = self.molecule1.HOMO_1 @ self.H_int @ self.molecule2.HOMO_1
        self.t_HOMO_2 = self.molecule1.HOMO_2 @ self.H_int @ self.molecule2.HOMO_2
        self.t_LUMO = self.molecule1.LUMO @ self.H_int @ self.molecule2.LUMO
        self.t_LUMO_1 = self.molecule1.LUMO_1 @ self.H_int @ self.molecule2.LUMO_1
        self.t_LUMO_2 = self.molecule1.LUMO_2 @ self.H_int @ self.molecule2.LUMO_2

    def __repr__(self):
        return f'BasePair({self.molecule1}, {self.molecule2}, identifier = "{self.identifier}")'

    def save_results(self, directory="results"):
        """
        Save the results of the calculation to a JSON file.

        Parameters
        ----------
        directory : str, optional
            The directory where the results file will be saved (default is "results").

        Notes
        -----
        .. note::

            The results include the HOMO and LUMO energies of two molecules and their coupling terms.
            The results are saved in a JSON file named after the identifier of the object.
        """

        dictionary = {
            "E1_HOMO": round(float(self.molecule1.E_HOMO), 4),
            "E2_HOMO": round(float(self.molecule2.E_HOMO), 4),
            "t_HOMO": round(float(self.t_HOMO), 4),
            "E1_LUMO": round(float(self.molecule1.E_LUMO), 4),
            "E2_LUMO": round(float(self.molecule2.E_LUMO), 4),
            "t_LUMO": round(float(self.t_LUMO), 4),
        }

        filename = self.identifier + ".json"
        filepath = os.path.join(directory, filename)

        with open(filepath, "w", encoding="utf-8") as f:
            json.dump(dictionary, f, indent=2)
        print("Results saved at" + filepath)

    def _calc_H_int(self):
        H_int = np.zeros((self.molecule1.num_orbitals, self.molecule2.num_orbitals))
        for i, orbital1 in enumerate(self.molecule1.orbitals):
            for j, orbital2 in enumerate(self.molecule2.orbitals):
                value = calc_orbital_overlap(orbital1, orbital2, "interbase")
                H_int[i, j] = value
        return H_int



"""This module provides functions to generate configurations for different tight-binding
models.

Shortcuts
---------
- WM: Wire Model
- LM: Ladder Model
- ELM: Extended Ladder Model
- F: Fishbone
- FC: Fully Connected
"""

from .. import TB_MODELS_PROPS
from ..tools import TB_MODELS

__all__ = ["get_tb_config", "TB_MODELS_PROPS"]

# --------------------------------------------------------------------------


def get_wm_config(num_sites_per_strand, strand=0, reversed_direction=False):
    """Generate the configuration for the Wire Model (WM).

    Parameters
    ----------
    num_sites_per_strand : int
        The number of sites per strand.
    strand : int, optional
        The strand index, by default 0.
    reversed_direction : bool, optional
        If True, reverses the direction of the connections, by default False.

    Returns
    -------
    list of tuple
        The configuration of the Wire Model.
    """

    e_list = [
        ("E", str((strand, tb_site)), str((strand, tb_site)))
        for tb_site in range(num_sites_per_strand)
    ]
    t_list = [
        (
            "t",
            str((strand, tb_site + (1 if not reversed_direction else 0))),
            str((strand, tb_site + (0 if not reversed_direction else 1))),
        )
        for tb_site in range(num_sites_per_strand - 1)
    ]
    return e_list + t_list


def get_lm_config(num_sites_per_strand, strand1=0, strand2=1):
    """Generate the configuration for the Ladder Model (LM).

    Parameters
    ----------
    num_sites_per_strand : int
        The number of sites per strand.
    strand1 : int, optional
        The first strand index, by default 0.
    strand2 : int, optional
        The second strand index, by default 1.

    Returns
    -------
    list of tuple
        The configuration of the Ladder Model.
    """

    wm_config_strand1 = get_wm_config(num_sites_per_strand, strand=strand1)
    wm_config_strand2 = get_wm_config(
        num_sites_per_strand, strand=strand2, reversed_direction=True
    )
    h_list = [
        ("h", str((strand1, site)), str((strand2, site)))
        for site in range(num_sites_per_strand)
    ]
    return wm_config_strand1 + wm_config_strand2 + h_list


def get_elm_config(num_sites_per_strand, strand1=0, strand2=1):
    """Generate the configuration for the Extended Ladder Model (ELM).

    Parameters
    ----------
    num_sites_per_strand : int
        The number of sites per strand.
    strand1 : int, optional
        The first strand index, by default 0.
    strand2 : int, optional
        The second strand index, by default 1.

    Returns
    -------
    list of tuple
        The configuration of the Extended Ladder Model.
    """

    lm_config = get_lm_config(num_sites_per_strand, strand1=strand1, strand2=strand2)
    r_plus_list = [
        ("r+", str((strand1, site)), str((strand2, site + 1)))
        for site in range(num_sites_per_strand - 1)
    ]
    r_minus_list = [
        ("r-", str((strand1, site + 1)), str((strand2, site)))
        for site in range(num_sites_per_strand - 1)
    ]
    return lm_config + r_plus_list + r_minus_list


def get_f_config(num_sites_per_strand, strand1=0, strand2=2):
    """Generate the configuration for the Fishbone Model (F).

    Parameters
    ----------
    num_sites_per_strand : int
        The number of sites per strand.
    strand1 : int, optional
        The first strand index, by default 0.
    strand2 : int, optional
        The second strand index, by default 2.

    Returns
    -------
    list of tuple
        The configuration of the Fishbone Model.
    """

    e_list = [
        ("E", str((strand1, tb_site)), str((strand1, tb_site)))
        for tb_site in range(num_sites_per_strand)
    ]
    e_list += [
        ("E", str((strand2, tb_site)), str((strand2, tb_site)))
        for tb_site in range(num_sites_per_strand)
    ]
    h_list1 = [
        ("h", str((strand1, tb_site)), str((strand1 + 1, tb_site)))
        for tb_site in range(num_sites_per_strand)
    ]
    h_list2 = [
        ("h", str((strand2 - 1, tb_site)), str((strand2, tb_site)))
        for tb_site in range(num_sites_per_strand)
    ]
    return e_list + h_list1 + h_list2


def get_fwm_config(num_sites_per_strand):
    """Generate the configuration for the Fishbone Wire Model (FWM).

    Parameters
    ----------
    num_sites_per_strand : int
        The number of sites per strand.

    Returns
    -------
    list of tuple
        The configuration of the Fishbone Wire Model.
    """
    f_config = get_f_config(num_sites_per_strand, strand1=0, strand2=2)
    wm_config = get_wm_config(num_sites_per_strand, strand=1)
    return f_config + wm_config


def get_flm_config(num_sites_per_strand):
    """Generate the configuration for the Fishbone Ladder Model (FLM).

    Parameters
    ----------
    num_sites_per_strand : int
        The number of sites per strand.

    Returns
    -------
    list of tuple
        The configuration of the Fishbone Ladder Model.
    """
    f_config = get_f_config(num_sites_per_strand, strand1=0, strand2=3)
    lm_config = get_lm_config(num_sites_per_strand, strand1=1, strand2=2)
    return f_config + lm_config


def get_felm_config(num_sites_per_strand):
    """Generate the configuration for the Fishbone Extended Ladder Model (FELM).

    Parameters
    ----------
    num_sites_per_strand : int
        The number of sites per strand.

    Returns
    -------
    list of tuple
        The configuration of the Fishbone Extended Ladder Model.
    """
    f_config = get_f_config(num_sites_per_strand, strand1=0, strand2=3)
    elm_config = get_elm_config(num_sites_per_strand, strand1=1, strand2=2)
    return f_config + elm_config


def get_fc_config(num_sites_per_strand):
    """Generate the configuration for the Fully Connected Model (FC).

    Parameters
    ----------
    num_sites_per_strand : int
        The number of sites per strand.

    Returns
    -------
    list of tuple
        The configuration of the Fully Connected Model.
    """
    felm_config = get_felm_config(num_sites_per_strand)
    t_list1 = [
        ("t", str((0, tb_site)), str((0, tb_site + 1)))
        for tb_site in range(num_sites_per_strand - 1)
    ]
    t_list2 = [
        ("t", str((3, tb_site)), str((3, tb_site + 1)))
        for tb_site in range(num_sites_per_strand - 1)
    ]
    return felm_config + t_list1 + t_list2


# --------------------------------------------------------------

TB_CONFIGS = dict(
    zip(
        TB_MODELS,
        [
            get_wm_config,
            get_lm_config,
            get_elm_config,
            get_fwm_config,
            get_flm_config,
            get_felm_config,
            get_fc_config,
        ],
    )
)


def get_tb_config(tb_model_name, tb_dims):
    """Get the tight-binding configuration for a specified model.

    Parameters
    ----------
    tb_model_name : str
        The name of the tight-binding model.
    tb_dims : tuple
        The dimensions of the model (number of strands, number of sites per strand).

    Returns
    -------
    list of tuple
        The configuration of the specified tight-binding model.

    Examples
    --------
    >>> get_tb_config("WM", (1, 3))
    [('E', '(0, 0)', '(0, 0)'),
    ('E', '(0, 1)', '(0, 1)'),
    ('E', '(0, 2)', '(0, 2)'),
    ('t', '(0, 1)', '(0, 0)'),
    ('t', '(0, 2)', '(0, 1)')]
    """

    tb_model_name = tb_model_name.upper()
    _, num_sites_per_strand = tb_dims
    return TB_CONFIGS[tb_model_name](num_sites_per_strand)



"""This module provides utility functions for unit conversion, particularly for
converting physical quantities to different units of measurement.

It includes functions to convert charge separation to dipole moments, convert between
various units, generate all possible conversion factors, and convert all values in a
dictionary from one unit to another.
"""

from itertools import permutations

import numpy as np
import scipy.constants as c

from ..tools import UNITS

__all__ = [
    "UNITS",
    "get_conversion",
    "get_all_conversions",
    "get_conversion_dict",
    "convert_to_debye",
]

# ------------------------------------------------


def convert_to_debye(charge_separation):
    """Converts the charge separation of two particles with elementary charge in
    Angstrom to a dipole moment in Debye.

    Parameters
    ----------
    charge_separation: float
        The electron and hole separation in Angstrom.

    Returns
    -------
    float
        The electrical dipole moment in Debye.
    """

    dipole_moment = c.e * 1e-10 * charge_separation
    return 100 * c.c * dipole_moment / 1e-19


def get_conversion(start_unit, end_unit):
    """Converts a value from the start_unit to the end_unit.

    Parameters
    ----------
    start_unit : str
        The unit to convert from.
    end_unit : str
        The unit to convert to.

    Returns
    -------
    float
        The conversion factor between the two units.

    Examples
    --------
    >>> 1/get_conversion("100meV", "rad/fs").
    """

    # Conversion factors to Joule
    convert_to_Joule = {
        "J": 1,
        "eV": c.e,
        "meV": 1e-3 * c.e,
        "100meV": 100 * 1e-3 * c.e,
        "rad/fs": 1 / 1e-15 * c.hbar,
        "rad/ps": 1 / 1e-12 * c.hbar,
        "1/cm": c.c / 1e-2 * 2 * np.pi * c.hbar,
        "K": c.k,
        "1/fs": 2 * np.pi / 1e-15 * c.hbar,
        "1/ps": 2 * np.pi / 1e-12 * c.hbar,
    }

    # Check if the units are defined
    if start_unit not in UNITS:
        raise ValueError(f"Unknown unit: {start_unit}. Defined units: {UNITS}")
    if end_unit not in UNITS:
        raise ValueError(f"Unknown unit: {end_unit}. Defined units: {UNITS}")

    # Return the conversion factor
    return convert_to_Joule[start_unit] / convert_to_Joule[end_unit]


def get_all_conversions():
    """Generates a dictionary with all possible conversions between units.

    Returns
    -------
    Dict[str, float]
        A dictionary where the keys are conversion descriptions
        and the values are the conversion factors.
    """

    conversion_dict = {}
    for start_unit, end_unit in permutations(UNITS, 2):
        conversion = start_unit + "_to_" + end_unit
        conversion_val = get_conversion(start_unit, end_unit)
        conversion_dict[conversion] = conversion_val
    return conversion_dict


def get_conversion_dict(param_dict, start_unit, end_unit):
    """Convert the values in a dictionary from one unit to another.

    Parameters
    ----------
    param_dict : dict
        Dictionary containing the parameters to be converted. Keys are parameter names and values are the parameter values.
    start_unit : str
        The unit of the input values in `param_dict`.
    end_unit : str
        The unit to which the values in `param_dict` should be converted.

    Returns
    -------
    dict
        A new dictionary with the same keys as `param_dict`, but with values converted to `end_unit`.
    """

    conversion_val = get_conversion(start_unit, end_unit)
    converted_dict = {key: value * conversion_val for key, value in param_dict.items()}
    return converted_dict



"""This module provides classes and functions to construct Lindblad dissipators for
master equations, specifically for quantum DNA models.

Shortcuts
---------
diss: dissipator
loc: local
glob: global
deph: dephasing
therm: thermalizing
reorg: reorganization
eigv: eigenvalue
eigs: eigensystem
pop: population
"""

from itertools import product
import copy

import numpy as np
import qutip as q

from ..tools import DEFAULTS, UNITS, check_diss_kwargs
from ..utils import get_conversion
from ..hamiltonian import TB_Ham, add_groundstate

from .relax_ops import get_relax_ops
from .deph_ops import (
    get_loc_deph_ops,
    get_loc_deph_p_ops,
    get_glob_deph_ops,
    get_glob_deph_p_ops,
)
from .therm_ops import get_loc_therm_ops, get_glob_therm_ops
from .observables import get_eh_observable

__all__ = ["Lindblad_Diss"]

# -------------------------------------- Lindblad operators -------------------------------------


class Lindblad_Diss:
    r"""Provides the operators of the form :math:`\sqrt{\gamma} a` that are used in the
    Lindblad dissipator of the master equation.

    Parameters
    ----------
    tb_ham : TBHamType
        The tight-binding Hamiltonian.
    diss_kwargs : dict
        Additional keyword arguments for the dissipator.

    Attributes
    ----------
    tb_ham : TBHamType
        The tight-binding Hamiltonian.
    num_sites : int
        Number of sites in the tight-binding model.
    loc_deph_rate : float
        Local dephasing rate.
    glob_deph_rate : float
        Global dephasing rate.
    uniform_relaxation : bool
        Flag for uniform relaxation.
    relax_rates : dict
        Relaxation rates for each DNA base.
    loc_therm : bool
        Flag for local thermalization.
    glob_therm : bool
        Flag for global thermalization.
    deph_rate : float
        Dephasing rate.
    cutoff_freq : float
        Cutoff frequency.
    reorg_energy : float
        Reorganization energy.
    temperature : float
        Temperature in Kelvin.
    spectral_density : str
        Type of spectral density ('debye' or 'ohmic').
    exponent : float
        Exponent for the Ohmic spectral density.
    relax_ops : list
        List of relaxation operators.
    deph_ops : list
        List of dephasing operators.
    therm_ops : list
        List of thermalizing operators.
    c_ops : list
        List of collapse operators.
    num_c_ops : int
        Number of collapse operators.
    e_ops : tuple
        Tuple containing population, coherence, and ground state population operators.
    pop_ops : dict
        Population operators.
    coh_ops : dict
        Coherence operators.
    groundstate_pop_ops : dict
        Ground state population operators.
    unit : str
        Unit of the operators.
    """

    def __init__(self, tb_ham, **diss_kwargs):
        # Check inputs
        assert isinstance(
            tb_ham, TB_Ham
        ), "tb_ham must be an instance of the class TB_Ham"
        self.diss_kwargs = copy.copy(DEFAULTS["diss_kwargs_default"])
        self.diss_kwargs.update(diss_kwargs)
        check_diss_kwargs(**self.diss_kwargs)
        self.verbose = DEFAULTS["verbose"]

        if self.verbose:
            print("Successfully checked all inputs for the Lindblad_Diss instance.")

        # Tight-binding Hamiltonian
        self.tb_ham = tb_ham
        self.num_sites = self.tb_ham.tb_model.num_sites
        self._unit = self.tb_ham.unit

        # Dephasing
        self.loc_deph_rate = self.diss_kwargs.get("loc_deph_rate")
        self.glob_deph_rate = self.diss_kwargs.get("glob_deph_rate")

        # Dephasing operators
        self.deph_ops = self._get_deph_ops()

        # Relaxation
        self.uniform_relaxation = self.diss_kwargs.get("uniform_relaxation")
        if self.uniform_relaxation:
            relax_rate = self.diss_kwargs["relax_rate"]
            tb_sites = tb_ham.tb_sites_flattened
            self.relax_rates = dict(zip(tb_sites, [relax_rate] * len(tb_sites)))
        else:
            self.relax_rates = self.diss_kwargs.get("relax_rates")
            assert set(self.relax_rates.keys()) == set(
                tb_ham.tb_sites_flattened
            ), "relax_rates must have the same keys as the tight-binding sites"

        # Relaxation operators
        self.relax_ops = self._get_relax_ops()

        # Thermalization
        self.loc_therm = self.diss_kwargs.get("loc_therm")
        self.glob_therm = self.diss_kwargs.get("glob_therm")
        self.deph_rate = self.diss_kwargs.get("deph_rate")
        self.cutoff_freq = self.diss_kwargs.get("cutoff_freq")
        self.reorg_energy = self.diss_kwargs.get("reorg_energy")
        self.temperature = self.diss_kwargs.get("temperature")
        self.spectral_density = self.diss_kwargs.get("spectral_density")
        self.exponent = self.diss_kwargs.get("exponent")

        # Thermalization operators
        self.therm_ops = self._get_therm_ops()

        # Collapse operators
        self._c_ops = self.relax_ops + self.deph_ops + self.therm_ops
        self.num_c_ops = len(self.c_ops)

        # Observables: population, coherence, and ground state population operators
        if tb_ham.description == "2P":
            self.e_ops = self._get_e_ops()
            self.pop_ops, self.coh_ops, self.groundstate_pop_ops = self.e_ops

        if self.verbose:
            print("Successfully initialized the Lindblad_Diss instance.")

    def __vars__(self):
        """Returns the instance variables as a dictionary."""
        return vars(self)

    def __repr__(self):
        """Returns a string representation of the Lindblad_Diss instance."""
        return f"Lindblad_Diss({self.tb_ham}, {self.diss_kwargs})"

    def __eq__(self, other):
        """Compares two Lindblad_Diss instances for equality."""
        return self.__repr__() == other.__repr__()

    # --------------------------------------------------------------------

    @property
    def c_ops(self):  # pylint: disable=missing-function-docstring
        return self._c_ops

    @c_ops.setter
    def c_ops(self, new_c_ops):
        assert isinstance(new_c_ops, list), "new_c_ops must be of type list"
        old_c_ops = self._c_ops
        self._c_ops = new_c_ops

        # Update the number of collapse operators
        if new_c_ops != old_c_ops:
            self.num_c_ops = len(self.c_ops)

    @property
    def unit(self):  # pylint: disable=missing-function-docstring
        return self._unit

    @unit.setter
    def unit(self, new_unit):
        assert isinstance(new_unit, str), "new_unit must be of type str"
        assert new_unit in UNITS, f"new_unit must be in {UNITS}"
        old_unit = self._unit
        self._unit = new_unit

        # Update the collapse operators
        if new_unit != old_unit:
            self.c_ops = [
                c_op * np.sqrt(get_conversion(old_unit, new_unit))
                for c_op in self.c_ops
            ]

    # -------------------------------------------------------------------

    def _get_relax_ops(self):
        """Generate relaxation operators based on relaxation rates and Hamiltonian.

        Returns
        -------
        list
            A list of relaxation operators, each scaled by the square root of the corresponding relaxation rate.
        """

        if not self.tb_ham.relaxation:
            return []

        tb_basis = self.tb_ham.tb_basis
        tb_basis_sites_dict = self.tb_ham.tb_basis_sites_dict

        return get_relax_ops(tb_basis, tb_basis_sites_dict, self.relax_rates)

    def _get_deph_ops(self):
        """Generate dephasing operators based on the system's Hamiltonian description
        and dephasing rates. This method computes local and global dephasing operators
        for both one-particle (1P) and two-particle (2P) descriptions of the system. The
        dephasing operators are determined by the eigensystem of the Hamiltonian and the
        specified local and global dephasing rates.

        Returns
        -------
        deph_p_ops : list
            A list of dephasing operators, scaled by the square root of the corresponding dephasing rate.
            The length and content of the list depend on the Hamiltonian description and the dephasing rates:
            - For local dephasing in a 2P description: 2 * num_sites operators.
            - For local dephasing in a 1P description: num_sites operators.
            - For global dephasing in a 2P description: num_sites^2 operators.
            - For global dephasing in a 1P description: num_sites operators.

        Raises
        ------
        AssertionError
            If the number of generated dephasing operators does not match the expected number
            based on the Hamiltonian description and the number of sites.
        """

        _, eigs = self.tb_ham.get_eigensystem()
        deph_ops = []

        # Local dephasing operators
        if self.loc_deph_rate:
            # Local dephasing operators for two particle description
            if self.tb_ham.description == "2P":
                loc_deph_ops = get_loc_deph_ops(
                    self.tb_ham.tb_basis, self.loc_deph_rate, self.tb_ham.relaxation
                )
                assert len(loc_deph_ops) == 2 * self.num_sites
                deph_ops = loc_deph_ops

            # Local dephasing operators for one particle description
            if self.tb_ham.description == "1P":
                loc_deph_p_ops = get_loc_deph_p_ops(
                    self.tb_ham.tb_basis, self.loc_deph_rate
                )
                assert len(loc_deph_p_ops) == self.num_sites
                deph_ops = loc_deph_p_ops

        # Global dephasing operators
        if self.glob_deph_rate:
            # Global dephasing operators for two particle description
            if self.tb_ham.description == "2P":
                glob_deph_ops = get_glob_deph_ops(
                    eigs, self.glob_deph_rate, self.tb_ham.relaxation
                )
                assert len(glob_deph_ops) == self.num_sites**2
                deph_ops = glob_deph_ops

            # Global dephasing operators for one particle description
            if self.tb_ham.description == "1P":
                glob_deph_p_ops = get_glob_deph_p_ops(eigs, self.glob_deph_rate)
                assert len(glob_deph_p_ops) == self.num_sites
                deph_ops = glob_deph_p_ops

        return deph_ops

    def _get_therm_ops(self):
        """Generate thermal operations based on the eigensystem of the Hamiltonian. This
        method calculates the thermal operations using either local or global
        thermalization processes, depending on the configuration of the instance.

        Returns
        -------
        therm_ops : list
            A list of thermal operations, scaled by the square root of the corresponding thermalization rate.

        Notes
        -----
        - The method first retrieves the eigensystem of the Hamiltonian and converts
          the eigenvalues to the desired units.
        - If `loc_therm` is set to True, local thermal operations are calculated.
        - If `glob_therm` is set to True, global thermal operations are calculated.
        - The thermal operations are determined by various parameters such as
          relaxation, dephasing rate, cutoff frequency, reorganization energy,
          temperature, spectral density, and exponent.
        """

        eigv, eigs = self.tb_ham.get_eigensystem()
        # Important: all parameters must be in the same units as the Hamiltonian.
        # If you want to give them in "rad/ps" you should uncomment the following line.
        # eigv *= get_conversion(self.tb_ham.unit, "rad/ps")
        therm_ops = []

        # Local thermalizing operators
        if self.loc_therm:
            loc_therm_ops = get_loc_therm_ops(
                eigv,
                eigs,
                self.tb_ham.relaxation,
                deph_rate=self.deph_rate,
                cutoff_freq=self.cutoff_freq,
                reorg_energy=self.reorg_energy,
                temperature=self.temperature,
                spectral_density=self.spectral_density,
                exponent=self.exponent,
            )
            therm_ops = loc_therm_ops

        # Global thermalizing operators
        if self.glob_therm:
            glob_therm_ops = get_glob_therm_ops(
                eigv,
                eigs,
                self.tb_ham.relaxation,
                deph_rate=self.deph_rate,
                cutoff_freq=self.cutoff_freq,
                reorg_energy=self.reorg_energy,
                temperature=self.temperature,
                spectral_density=self.spectral_density,
                exponent=self.exponent,
            )
            therm_ops = glob_therm_ops

        return therm_ops

    def _get_e_ops(self):
        """Generates the expectation value operators (e_ops) for the Lindblad master
        equation. This method constructs dictionaries of population and coherence
        operators for each particle in the tight-binding Hamiltonian (tb_ham). It also
        includes ground state population operators if relaxation is considered.

        Returns
        -------
        tuple of dict
            A tuple containing three dictionaries:
            - pop_dict: Population operators for each particle and site.
            - coh_dict: Coherence operators for each particle and pair of sites.
            - groundstate_pop_dict: Ground state population operators (only if relaxation is considered).

        Raises
        ------
        AssertionError
            If the Hamiltonian description is not "2P". This method is only defined for the "2P" description.
        """

        assert self.tb_ham.description == "2P", (
            "Only defined for 2P description. To treat the 1P case it is more efficient to "
            "return the whole density matrices and work with it."
        )

        pop_dict, coh_dict, groundstate_pop_dict = {}, {}, {}

        # Population and coherence operators
        for particle in self.tb_ham.particles:
            for tb_site1, tb_site2 in product(self.tb_ham.tb_basis, repeat=2):
                # Create electron-hole observable
                observable = get_eh_observable(
                    self.tb_ham.tb_basis, particle, tb_site1, tb_site2
                )
                if self.tb_ham.relaxation:
                    observable = add_groundstate(observable)

                # Add observable as population operator
                if tb_site1 == tb_site2:
                    key = particle + "_" + tb_site1
                    pop_dict[key] = q.Qobj(observable)

                # Add observable as coherence operator
                else:
                    key = particle + "_" + tb_site1 + "_" + tb_site2
                    coh_dict[key] = q.Qobj(observable)

        # Ground state population operator
        if self.tb_ham.relaxation:
            groundstate_pop_dict["groundstate"] = q.fock_dm(self.tb_ham.matrix_dim, 0)

        return pop_dict, coh_dict, groundstate_pop_dict



# pylint: skip-file

import sys
import customtkinter as ctk
from tkinter.scrolledtext import ScrolledText
import re

# ------------------------------------------------------------

import sys
import customtkinter as ctk
from tkinter.scrolledtext import ScrolledText


class RedirectText:
    def __init__(self, text_widget):
        self.output = text_widget

    def write(self, string):
        """Handles both stdout and stderr redirection."""
        self.output.configure(state="normal")
        self.output.insert(ctk.END, string.strip() + "\n")  # Ensures new line output
        self.output.see(ctk.END)
        self.output.configure(state="disabled")

    def flush(self):
        pass  # Required for compatibility with sys.stdout and sys.stderr


class ScrollableConsoleFrame(ctk.CTkFrame):
    def __init__(self, master=None, **kwargs):
        super().__init__(master, **kwargs)

        # Create a ScrolledText widget for the console output
        self.console_output = ScrolledText(
            self, wrap="word", state="disabled", height=10, width=35
        )
        self.console_output.grid(row=0, column=0, sticky="nsew", padx=(10, 0), pady=10)

        scrollbar = ctk.CTkScrollbar(self, command=self.console_output.yview)
        scrollbar.grid(row=0, column=1, sticky="ns", padx=(0, 10), pady=10)
        self.console_output["yscrollcommand"] = scrollbar.set

        # Redirect both stdout and stderr to the ScrolledText widget
        self.redirected_output = RedirectText(self.console_output)
        sys.stdout = self.redirected_output
        sys.stderr = self.redirected_output

        # Override sys.excepthook to filter AssertionError traceback
        sys.excepthook = self.handle_exceptions

        # Override Tkinter's error handling to capture GUI-related assertions
        self.master.report_callback_exception = self.handle_tkinter_exceptions

    def handle_exceptions(self, exc_type, exc_value, exc_traceback):
        """Custom exception handler that filters out assertion tracebacks."""
        if issubclass(exc_type, AssertionError):
            sys.stderr.write(str(exc_value) + "\n")  # Only print the assertion message
        else:
            import traceback

            sys.stderr.write(
                "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
            )  # Keep full error

    def handle_tkinter_exceptions(self, exc_type, exc_value, exc_traceback):
        """Handles Tkinter's internal exceptions to prevent unwanted traceback."""
        if issubclass(exc_type, AssertionError):
            sys.stderr.write(
                "Warning: " + str(exc_value) + "\n-------------------------------"
            )  # Print only assertion message
        else:
            import traceback

            sys.stderr.write(
                "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
            )  # Full traceback



"""Module for reducing density matrices to the electron, hole or exciton subspace."""

from itertools import product
import numpy as np

from ..hamiltonian import delete_groundstate
from ..environment import get_eh_observable, get_tb_observable

__all__ = ["get_reduced_dm", "get_reduced_dm_eigs"]

# ------------------------------------------------


def get_reduced_dm(dm, particle, tb_basis):
    """Reduces the density matrix for a specific particle type.

    Parameters
    ----------
    dm : np.ndarray
        The initial density matrix.
    particle : str
        The type of particle ('electron' or 'hole').
    tb_basis : List[str]
        The list of tight-binding site basis states.

    Returns
    -------
    np.ndarray
        The reduced density matrix for the specified particle.

    Raises
    ------
    ValueError
        If the particle type is not recognized.

    Examples
    --------
    >>> dm = np.eye(4)
    >>> get_reduced_dm(dm, 'electron', ['(0, 0)', '(1, 0)'])
    array([[2., 0.],
           [0., 2.]])
    """

    num_sites = len(tb_basis)

    # Before taking the trace the groundstate needs to be removed
    if dm.shape[0] != num_sites**2:
        dm = delete_groundstate(dm)

    reduced_dm = np.zeros((num_sites, num_sites), dtype=complex)
    for start_state, end_state in product(tb_basis, repeat=2):
        # Calculate expectation using the full density matrix
        observable = get_eh_observable(tb_basis, particle, start_state, end_state)
        value = np.trace(observable @ dm)

        # Multiply expectation value with the corresponding element of the reduced density matrix
        reduced_dm += value * get_tb_observable(tb_basis, start_state, end_state).T
    return reduced_dm


def get_reduced_dm_eigs(tb_ham, particle, eigenstate_idx):
    """Reduces the density matrix of a selected eigenstate of the Hamiltonian for a
    specific particle type.

    Parameters
    ----------
    tb_ham : TBHamType
        The Hamiltonian object containing the matrix and site basis.
    particle : str
        The type of particle ('electron' or 'hole').
    eigenstate_idx : int
        The index of the eigenstate.

    Returns
    -------
    np.ndarray
        The reduced density matrix.
    """

    _, eigs = tb_ham.get_eigensystem()
    dm = np.outer(eigs[:, eigenstate_idx], eigs[:, eigenstate_idx].conj())
    return get_reduced_dm(dm, particle, tb_ham.tb_basis)



import pytest

from qDNA import DNA_Seq, create_upper_strands


@pytest.mark.parametrize(
    "seq, mode, methylated, lower_strand, expected",
    [
        (list("GCG"), "ELM", True, "auto_complete", (list("GCG"), list("CGC"))),
        (list("GCG"), "WM", True, "auto_complete", (list("GCG"),)),
        (
            list("GCG"),
            "FLM",
            True,
            "auto_complete",
            (list("BBB"), list("GCG"), list("CGC"), list("BBB")),
        ),
        (list("GCGCG"), "LM", False, list("FGFGC"), (list("GCGCG"), list("FGFGC"))),
    ],
)
def test_DNA_Seq(seq, mode, methylated, lower_strand, expected):
    assert (
        DNA_Seq(seq, mode, methylated=methylated, lower_strand=lower_strand).dna_seq
        == expected
    )


@pytest.mark.parametrize(
    "length, bases, expected",
    [(3, ["A", "T"], ["AAA", "AAT", "ATA", "ATT", "TAA", "TAT", "TTA", "TTT"])],
)
def test_create_upper_strands(length, bases, expected):
    assert create_upper_strands(length, bases) == expected



import matplotlib as mpl

from ..tools import DNA_BASES, PARTICLES

mpl.rcdefaults()
try:
    mpl.style.use("qDNA-default")
except OSError:
    print("Could not load qDNA-default style. Using seaborn-v0_8-paper style instead.")
    mpl.style.use("seaborn-v0_8-paper")

# A, T, G, C, F
COLORS_DNA_BASES = dict(
    zip(DNA_BASES, ["#a6cee3", "#1f78b4", "#b2df8a", "#33a02c"])
)  # seaborn color palette "Paired"
COLORS_DNA_BASES = dict(
    zip(DNA_BASES, ["#b93540", "#ed7e40", "#4167c7", "#60abcd", "#1f1e1e"])
)  # seaborn color palette "icefire" with 7 colors

COLORS_PARTICLES = dict(
    zip(PARTICLES, ["#4c72b0", "#55a868", "#c44e52"])
)  # seaborn color palette "deep"
COLORS_PARTICLES = dict(
    zip(PARTICLES, ["#4798ce", "#e25f33", "#1f1e1e"])
)  # seaborn color palette "icefire" with 5 colors

from .plot_dna_base_frequency import *
from .plot_eigenspectrum import *
from .plot_fourier import *
from .plot_pop import *



"""This module provides utility functions for analyzing density matrices.

The functions include calculations for trace distance, purity, coherence, and inverse
participation ratio (IPR).
"""

import numpy as np

__all__ = ["calc_trace_distance", "calc_purity", "calc_coherence", "calc_ipr_dm"]

# ------------------------------------------------


def calc_trace_distance(dm_1, dm_2):
    """Calculate the trace distance between two density matrices. The trace distance is
    a measure of the distinguishability between two quantum states represented by their
    density matrices. It is defined as half the trace of the absolute difference between
    the two density matrices.

    Parameters
    ----------
    dm_1 : numpy.ndarray
        The first density matrix.
    dm_2 : numpy.ndarray
        The second density matrix.

    Returns
    -------
    float
        The trace distance between the two density matrices.

    Raises
    ------
    ValueError
        If the density matrices do not have the same shape.
    """

    # Check if the density matrices have the same shape
    if dm_1.shape != dm_2.shape:
        raise ValueError("Density matrices must have the same shape.")

    # Calculate the trace distance
    eigenvals = np.linalg.eig(dm_1 - dm_2)[0]
    diag = np.diag(np.absolute(eigenvals))
    return 0.5 * np.trace(diag)


def calc_purity(dm):
    """Calculates the purity of a density matrix.

    Parameters
    ----------
    dm : np.ndarray
        Density matrix.

    Returns
    -------
    float
        The purity of the density matrix.
    """

    return np.real(np.trace(np.matmul(dm, dm)))


def calc_coherence(dm):
    """Calculates the coherence (absolute sum of the off-diagonals) of a density matrix.

    Parameters
    ----------
    dm : np.ndarray
        Density matrix.

    Returns
    -------
    float
        The coherence of the density matrix.
    """

    coherence = sum(abs(element) for element in dm.flatten()) - np.sum(np.diag(dm))
    return coherence.real


def calc_ipr_dm(dm):
    r"""Calculate the inverse participation ratio (IPR) of a density matrix.

    Parameters
    ----------
    dm : np.ndarray
        The density matrix for which the IPR is calculated. Should be a square, complex-valued matrix.

    Returns
    -------
    float
        The inverse participation ratio of the density matrix, providing an indication of state localization.

    Notes
    -----
    .. note::

        The IPR provides a measure of the localization of a quantum state

        .. math::

            \mathrm{IPR} = \frac{\left( \sum_{i,j} |\rho_{i,j}| \right)^2}{N \sum_{i,j} |\rho_{i,j}|^2}

        - Localized state: :math:`IPR = 1/N`, where `N` is the dimension of the density matrix.
        - Delocalized (coherent) state: :math:`IPR = N`, indicating full coherence.
        - Maximally mixed state: :math:`IPR = 1`, representing maximal mixing.
    """

    dims = dm.shape[0]
    numerator = sum(abs(element) for element in dm.flatten())
    denominator = sum(abs(element) ** 2 for element in dm.flatten())
    return numerator**2 / (dims * denominator)



# pylint: skip-file

import customtkinter as ctk


def change_state_all_widgets(frame, state):
    for widget in frame.winfo_children():
        if isinstance(
            widget, ctk.CTkBaseClass
        ):  # Check if the widget is a customtkinter widget
            widget.configure(state=state)
        elif isinstance(
            widget, ctk.CTkFrame
        ):  # Recursively disable widgets in nested frames
            change_state_all_widgets(widget, state)



"""This module provides functions to plot eigenenergies and eigenstates for quantum DNA
models."""

import numpy as np

from ..utils import get_conversion
from ..dynamics import get_reduced_dm_eigs

from . import PARTICLES, COLORS_PARTICLES

__all__ = ["PARTICLES", "COLORS_PARTICLES", "plot_eigv", "plot_eigs"]

# -------------------------------------------------------------------------------------


def plot_eigv(ax, tb_ham, energy_unit="100meV"):
    """Plots the eigenenergies.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot on.
    tb_ham : TB_Ham
        The tight-binding Hamiltonian.
    energy_unit : str, optional
        The unit of energy to plot, by default '100meV'.
    """

    # calculation
    eigv, _ = tb_ham.get_eigensystem()
    eigv *= get_conversion(tb_ham.unit, energy_unit)

    # plotting
    ax.set_title("Eigenvalues")
    ax.plot(eigv, "o")
    ax.set_ylabel("Energy in " + energy_unit)


def plot_eigs(ax, tb_ham, eigenstate_idx):
    """Plots the distribution of eigenstates over tight-binding sites.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot on.
    tb_ham : TB_Ham
        The tight-binding Hamiltonian.
    eigenstate_idx : int
        The index of the eigenstate to plot.
    """

    # calculation
    _, eigs = tb_ham.get_eigensystem()
    eigs_prob_distribution = {}

    # 1P description
    if tb_ham.description == "2P":
        for particle in tb_ham.particles:
            # extract eigenstate population for each local state
            reduced_dm_eigs = get_reduced_dm_eigs(tb_ham, particle, eigenstate_idx)
            eigs_prob_distribution[particle] = np.diag(reduced_dm_eigs).real

    # 2P description
    elif tb_ham.description == "1P":
        for particle in tb_ham.particles:
            # extract eigenstate population for each local state
            eigs_distribution = np.diag(eigs)
            eigs_prob_distribution[particle] = np.multiply(
                eigs_distribution, eigs_distribution.conj()
            ).real

    # plotting
    for particle in tb_ham.particles:
        ax.plot(
            tb_ham.tb_basis,
            eigs_prob_distribution[particle],
            label=particle,
            color=COLORS_PARTICLES[particle],
        )
    ax.set_ylim(0, 1.02)
    ax.set_title(f"Distribution of Eigenstate {eigenstate_idx}")
    ax.legend()



# pylint: skip-file

import customtkinter as ctk

# --------------------------------------------------


class DynamicsFrame(ctk.CTkFrame):
    def __init__(self, master, configs, **kwargs):
        """
        Notes:
            This frame is located in the options tab. This means that the options tab is its master.
        """

        # initialization of the ctk.CTkFrame class
        super().__init__(master)
        self.pack(fill="both", expand=True)
        self.configs = configs
        self.me_kwargs_default = self.configs["me_kwargs_default"]

        # widgets
        self.time_label = ctk.CTkLabel(
            self, text="Simulated Time", font=ctk.CTkFont(size=15, weight="bold")
        )
        self.time_label.grid(row=0, column=0, pady=10, padx=10)

        self.me_t_steps_label = ctk.CTkLabel(self, text="Time Steps:")
        self.me_t_steps_label.grid(row=1, column=0, padx=10, pady=10)

        self.me_t_steps_entry = ctk.CTkEntry(self)
        self.me_t_steps_entry.insert(0, str(self.me_kwargs_default["t_steps"]))
        self.me_t_steps_entry.grid(row=1, column=1, padx=10, pady=10)

        self.me_t_end_label = ctk.CTkLabel(self, text="Time End:")
        self.me_t_end_label.grid(row=2, column=0, padx=10, pady=10)

        self.me_t_end_entry = ctk.CTkEntry(self)
        self.me_t_end_entry.insert(0, str(self.me_kwargs_default["t_end"]))
        self.me_t_end_entry.grid(row=2, column=1, padx=10, pady=10)

        self.me_t_unit_label = ctk.CTkLabel(self, text="Time Unit:")
        self.me_t_unit_label.grid(row=3, column=0, padx=10, pady=10)

        self.me_t_unit_combo = ctk.CTkComboBox(self, values=self.configs["T_UNITS"])
        self.me_t_unit_combo.set(self.me_kwargs_default["t_unit"])
        self.me_t_unit_combo.grid(row=3, column=1, padx=10, pady=10)

        self.initial_state_label = ctk.CTkLabel(
            self, text="Initial State", font=ctk.CTkFont(size=15, weight="bold")
        )
        self.initial_state_label.grid(row=4, column=0, pady=10, padx=10)

        self.me_init_e_state_label = ctk.CTkLabel(self, text="Initial Electron State:")
        self.me_init_e_state_label.grid(row=5, column=0, padx=10, pady=10)

        self.me_init_e_state_combo = ctk.CTkComboBox(
            self, values=self.me_kwargs_default["init_e_state"]
        )
        self.me_init_e_state_combo.grid(row=5, column=1, padx=10, pady=10)
        self.me_init_e_state_combo.set(self.me_kwargs_default["init_e_state"])

        self.me_init_h_state_label = ctk.CTkLabel(self, text="Initial Hole State:")
        self.me_init_h_state_label.grid(row=6, column=0, padx=10, pady=10)

        self.me_init_h_state_combo = ctk.CTkComboBox(
            self, values=self.me_kwargs_default["init_h_state"]
        )
        self.me_init_h_state_combo.grid(row=6, column=1, padx=10, pady=10)
        self.me_init_h_state_combo.set(self.me_kwargs_default["init_h_state"])

    def get_me_kwargs(self):
        """Returns the values of widgets with get() method in dictionary format."""

        me_kwargs = {
            "t_steps": float(self.me_t_steps_entry.get()),
            "t_end": float(self.me_t_end_entry.get()),
            "t_unit": self.me_t_unit_combo.get(),
            "init_e_state": self.me_init_e_state_combo.get(),
            "init_h_state": self.me_init_h_state_combo.get(),
        }
        return me_kwargs



"""This module provides functions to calculate the average exciton population for
quantum DNA models."""

import multiprocessing
from functools import partial

import numpy as np
from tqdm import tqdm

from ..dynamics import get_me_solver
from ..tools import load_json, save_json

__all__ = [
    "calc_backbone_transfer",
    "calc_average_transfer",
    "calc_exciton_transfer",
    "calc_exciton_transfer_wrapper",
    "calc_exciton_transfer_dict",
]

# ------------------------------------------------


def calc_average_transfer(tb_sites, me_solver, average=True):
    """Calculates the average populations on the given TB sites in a given time period.

    Parameters
    ----------
    tb_sites : List[str]
        The tight-binding sites for which the average population should be calculated.
    me_solver: MESolverType
        Instance of the ME_Solver class.
    average : bool
        Indicates if the exciton_population should be time-averaged.

    Returns
    -------
    dict
        The average populations for each particle.
    """
    average_pop = dict(
        zip(me_solver.tb_ham.particles, [0] * len(me_solver.tb_ham.particles))
    )
    for particle in me_solver.tb_ham.particles:
        avg_pop = np.sum(
            [me_solver.get_pop()[particle + "_" + tb_site] for tb_site in tb_sites],
            axis=0,
        )
        if average:
            avg_pop = np.mean(avg_pop)
        average_pop[particle] = avg_pop
    return average_pop


def calc_backbone_transfer(upper_strand, tb_model_name, **kwargs):
    """Calculates the average population of the backbone sites in a given time period.

    Parameters
    ----------
    upper_strand : str
        The upper strand of DNA sequence.
    tb_model_name : str
        The name of the tight-binding model.
    kwargs : dict
        Additional keyword arguments for the master equation solver.

    Returns
    -------
    dict
        The average backbone population for each particle.
    """
    me_solver = get_me_solver(upper_strand, tb_model_name, **kwargs)
    tb_model = me_solver.tb_ham.tb_model
    assert (
        tb_model.tb_model_name[0] == "F"
    ), "Backbone population can only be calculated for Fishbone models"
    upper_backbone_sites = [
        f"(0, {site})" for site in range(tb_model.num_sites_per_strand)
    ]
    lower_backbone_sites = [
        f"({tb_model.num_strands-1}, {site})"
        for site in range(tb_model.num_sites_per_strand)
    ]
    backbone_sites = upper_backbone_sites + lower_backbone_sites
    return calc_average_transfer(backbone_sites, me_solver)


def calc_exciton_transfer(upper_strand, tb_model_name, average=True, **kwargs):
    """Calculates the average exciton population on the upper and lower strand.

    Parameters
    ----------
    upper_strand : str
        The upper strand of DNA sequence.
    tb_model_name : str
        The name of the tight-binding model.
    average : bool
        Indicates if the exciton_population should be time-averaged.
    kwargs : dict
        Additional keyword arguments for the master equation solver.

    Returns
    -------
    Tuple[dict] or List[Tuple[dict]]
        The average exciton population on the upper and lower strand.

    Examples
    --------
    >>> calc_exciton_transfer("GCG", "ELM")
    ({'electron': 0.6867427675114343,
    'hole': 0.9943813264087192,
    'exciton': 0.45001514054414693},
    {'electron': 0.31325723248857257,
    'hole': 0.005618673591287626,
    'exciton': 0.0001960836601784245})
    """
    me_solver = get_me_solver(upper_strand, tb_model_name, **kwargs)
    tb_model = me_solver.tb_ham.tb_model
    assert tb_model.tb_model_name[0] != "F", "Not definded for Fishbone models"
    assert tb_model.tb_model_name[0] != "W", "Not definded for Wire models"

    upper_strand_sites = [
        f"(0, {j})" for j in range(me_solver.tb_model.num_sites_per_strand)
    ]
    lower_strand_sites = [
        f"(1, {j})" for j in range(me_solver.tb_model.num_sites_per_strand)
    ]
    upper_strand_pop = calc_average_transfer(
        upper_strand_sites, me_solver, average=average
    )
    lower_strand_pop = calc_average_transfer(
        lower_strand_sites, me_solver, average=average
    )
    return upper_strand_pop, lower_strand_pop


def calc_backbone_transfer_wrapper(
    upper_strand, tb_model_name, lifetime_dict, **kwargs
):
    """Calculates the average backbone population on the upper and lower strand.

    Parameters
    ----------
    upper_strand : str
        The upper strand of DNA sequence.
    tb_model_name : str
        The name of the tight-binding model.
    lifetime_dict : dict
        A dictionary containing the lifetimes for different DNA sequences.
    kwargs : dict
        Additional keyword arguments for the master equation solver.

    Returns
    -------
    dict
        The average backbone population on the upper and lower strand.
    """
    kwargs["t_end"] = lifetime_dict[upper_strand]
    kwargs["t_steps"] = kwargs["t_end"] // 2 + 2
    kwargs["t_unit"] = "fs"
    return calc_backbone_transfer(upper_strand, tb_model_name, **kwargs)


def calc_exciton_transfer_wrapper(upper_strand, tb_model_name, lifetime_dict, **kwargs):
    """Calculates the average exciton population on the upper and lower strand.

    Parameters
    ----------
    upper_strand : str
        The upper strand of DNA sequence.
    tb_model_name : str
        The name of the tight-binding model.
    lifetime_dict : dict
        A dictionary containing the lifetimes for different DNA sequences.
    kwargs : dict
        Additional keyword arguments for the master equation solver.

    Returns
    -------
    Tuple[dict] or List[Tuple[dict]]
        The average exciton population on the upper and lower strand.
    """
    kwargs["t_end"] = lifetime_dict[upper_strand]
    kwargs["t_steps"] = kwargs["t_end"] // 2 + 2
    kwargs["t_unit"] = "fs"
    return calc_exciton_transfer(upper_strand, tb_model_name, **kwargs)


def calc_exciton_transfer_dict(tb_model_name, filename, directory, num_cpu=None):
    """Calculates the average exciton population for multiple upper strands using
    multiprocessing.

    Parameters
    ----------
    tb_model_name : str
        The name of the tight-binding model.
    filename : str
        The filename to load the lifetime dictionary from.
    directory : str
        The directory where the lifetime dictionary is located.
    num_cpu : int, optional
        The number of CPU cores to use. Defaults to the total number of CPUs minus one.

    Returns
    -------
    Dict[str, float]
        Dictionary containing the average exciton population for each upper strand.
    """
    try:
        lifetime_dict, kwargs = load_json(
            "lifetime_" + filename,
            directory,
            load_metadata=True,
        )
    except FileNotFoundError as e:
        print(f"Could not load lifetime_dict: {e}")

    if not num_cpu:
        num_cpu = multiprocessing.cpu_count() - 1
    upper_strands = list(lifetime_dict.keys())
    partial_calc_exciton_transfer = partial(
        calc_exciton_transfer_wrapper,
        tb_model_name=tb_model_name,
        lifetime_dict=lifetime_dict,
        **kwargs,
    )
    with multiprocessing.Pool(processes=num_cpu) as pool:
        exciton_transfer_list = list(
            tqdm(
                pool.imap(partial_calc_exciton_transfer, upper_strands),
                total=len(upper_strands),
                # file=sys.stdout,
            )
        )

    exciton_transfer_dict = dict(zip(upper_strands, exciton_transfer_list))
    save_json(
        exciton_transfer_dict,
        kwargs,
        "exciton_transfer_" + filename,
        directory,
    )
    return exciton_transfer_dict



import os
import json
from itertools import combinations

import numpy as np

from .slater_koster import calc_orbital_overlap

# ----------------------------------------------


class BasePair:
    """
    A class to represent a base pair in a DNA molecule.

    Attributes
    ----------
    base1 : object
        The first base in the pair.
    base2 : object
        The second base in the pair.
    identifier : str, optional
        An identifier for the base pair (default is None).
    atoms : list
        List of atoms in the base pair.
    num_atoms : int
        Total number of atoms in the base pair.
    atom_identifiers : list
        List of atom identifiers in the base pair.
    atom_coordinates : ndarray
        Coordinates of atoms in the base pair.
    atom_distance_matrix : ndarray
        Matrix of distances between atoms in the base pair.
    orbitals : list
        List of orbitals in the base pair.
    num_orbitals : int
        Total number of orbitals in the base pair.
    orbital_identifiers : list
        List of orbital identifiers in the base pair.
    orbital_coordinates : ndarray
        Coordinates of orbitals in the base pair.
    orbital_distance_matrix : ndarray
        Matrix of distances between orbitals in the base pair.
    mass : float
        Mass of the base pair.
    center_of_mass : ndarray
        Center of mass of the base pair.
    relative_atom_coordinates : ndarray
        Atom coordinates relative to the center of mass.
    relative_orbital_coordinates : ndarray
        Orbital coordinates relative to the center of mass.
    num_electrons : int
        Total number of valence electrons in the base pair.
    H : ndarray
        LCAO Hamiltonian matrix for the base pair.
    eigv : ndarray
        Eigenvalues of the Hamiltonian matrix.
    eigs : ndarray
        Eigenvectors of the Hamiltonian matrix.
    E_HOMO : float
        Energy of the highest occupied molecular orbital (HOMO).
    HOMO : ndarray
        Highest occupied molecular orbital (HOMO).
    HOMO_occupation : ndarray
        Occupation of the HOMO.
    HOMO_type : str
        Type of the HOMO (sigma, pi, or n).
    HOMO_type_occupation : list
        Occupation of the HOMO by type.
    E_HOMO_1 : float
        Energy of the HOMO-1.
    HOMO_1 : ndarray
        HOMO-1.
    HOMO_occupation_1 : ndarray
        Occupation of the HOMO-1.
    HOMO_type_1 : str
        Type of the HOMO-1 (sigma, pi, or n).
    HOMO_type_occupation_1 : list
        Occupation of the HOMO-1 by type.
    E_HOMO_2 : float
        Energy of the HOMO-2.
    HOMO_2 : ndarray
        HOMO-2.
    HOMO_occupation_2 : ndarray
        Occupation of the HOMO-2.
    HOMO_type_2 : str
        Type of the HOMO-2 (sigma, pi, or n).
    HOMO_type_occupation_2 : list
        Occupation of the HOMO-2 by type.
    E_LUMO : float
        Energy of the lowest unoccupied molecular orbital (LUMO).
    LUMO : ndarray
        Lowest unoccupied molecular orbital (LUMO).
    LUMO_occupation : ndarray
        Occupation of the LUMO.
    LUMO_type : str
        Type of the LUMO (sigma, pi, or n).
    LUMO_type_occupation : list
        Occupation of the LUMO by type.
    E_LUMO_1 : float
        Energy of the LUMO+1.
    LUMO_1 : ndarray
        LUMO+1.
    LUMO_occupation_1 : ndarray
        Occupation of the LUMO+1.
    LUMO_type_1 : str
        Type of the LUMO+1 (sigma, pi, or n).
    LUMO_type_occupation_1 : list
        Occupation of the LUMO+1 by type.
    E_LUMO_2 : float
        Energy of the LUMO+2.
    LUMO_2 : ndarray
        LUMO+2.
    LUMO_occupation_2 : ndarray
        Occupation of the LUMO+2.
    LUMO_type_2 : str
        Type of the LUMO+2 (sigma, pi, or n).
    LUMO_type_occupation_2 : list
        Occupation of the LUMO+2 by type.
    MO_types : dict
        Dictionary of molecular orbital types.
    E_exc : float
        Excitation energy (HOMO-LUMO gap).

    Methods
    -------
    save_results(directory="results"):
        Saves the results to a JSON file.
    calc_MO(identifier, deviation=0):
        Returns properties of the molecular orbital (MO).
    """

    def __init__(self, base1, base2, identifier=None):
        self.base1 = base1
        self.base2 = base2

        # Identifier
        self.identifier = identifier
        if self.identifier is None:
            self.identifier = base1.identifier  # + "_" + base2.identifier

        # Atoms
        self.atoms = self.base1.atoms + self.base2.atoms
        self.num_atoms = self.base1.num_atoms + self.base2.num_atoms
        self.atom_identifiers = (
            self.base1.atom_identifiers + self.base2.atom_identifiers
        )
        self.atom_coordinates = np.block(
            [[self.base1.atom_coordinates], [self.base2.atom_coordinates]]
        )
        self.atom_distance_matrix = self._get_atom_distance_matrix()

        # Orbitals
        self.orbitals = self.base1.orbitals + self.base2.orbitals
        self.num_orbitals = self.base1.num_orbitals + self.base2.num_orbitals
        self.orbital_identifiers = (
            self.base1.orbital_identifiers + self.base2.orbital_identifiers
        )
        self.orbital_coordinates = np.block(
            [[self.base1.orbital_coordinates], [self.base2.orbital_coordinates]]
        )
        self.orbital_distance_matrix = self._get_orbital_distance_matrix()

        # Coordinates relative to the center of mass
        self.mass, self.center_of_mass = self._calc_center_of_mass()
        self.relative_atom_coordinates = np.array(
            [coordinates - self.center_of_mass for coordinates in self.atom_coordinates]
        )
        self.relative_orbital_coordinates = np.array(
            [
                coordinates - self.center_of_mass
                for coordinates in self.orbital_coordinates
            ]
        )

        # Valence electrons
        self.num_electrons = self.base1.num_electrons + self.base2.num_electrons

        # Molecular Orbitals
        # ------------------

        # LCAO Hamiltonian
        self.H = self._calc_H()

        # Solve the eigenvalue problem
        self.eigv, self.eigs = np.linalg.eigh(self.H)

        # HOMO and LUMO
        (
            self.E_HOMO,
            self.HOMO,
            self.HOMO_occupation,
            self.HOMO_type,
            self.HOMO_type_occupation,
        ) = self.calc_MO("HOMO")
        (
            self.E_HOMO_1,
            self.HOMO_1,
            self.HOMO_occupation_1,
            self.HOMO_type_1,
            self.HOMO_type_occupation_1,
        ) = self.calc_MO("HOMO", deviation=1)
        (
            self.E_HOMO_2,
            self.HOMO_2,
            self.HOMO_occupation_2,
            self.HOMO_type_2,
            self.HOMO_type_occupation_2,
        ) = self.calc_MO("HOMO", deviation=2)
        (
            self.E_LUMO,
            self.LUMO,
            self.LUMO_occupation,
            self.LUMO_type,
            self.LUMO_type_occupation,
        ) = self.calc_MO("LUMO")
        (
            self.E_LUMO_1,
            self.LUMO_1,
            self.LUMO_occupation_1,
            self.LUMO_type_1,
            self.LUMO_type_occupation_1,
        ) = self.calc_MO("LUMO", deviation=1)
        (
            self.E_LUMO_2,
            self.LUMO_2,
            self.LUMO_occupation_2,
            self.LUMO_type_2,
            self.LUMO_type_occupation_2,
        ) = self.calc_MO("LUMO", deviation=2)

        self.MO_types = {
            "LUMO_2": self.LUMO_type_2,
            "LUMO_1": self.LUMO_type_1,
            "LUMO": self.LUMO_type,
            "HOMO": self.HOMO_type,
            "HOMO_1": self.HOMO_type_1,
            "HOMO_2": self.HOMO_type_2,
        }

        # Excitation energy (HOMO-LUMO gap)
        self.E_exc = self.E_LUMO - self.E_HOMO

    # ---------------------------------------------------------

    def __repr__(self):
        return f'BasePair({self.base1}, {self.base2}, identifier = "{self.identifier}")'

    def save_results(self, directory="results"):
        """
        Save the calculation results to a JSON file.

        Parameters
        ----------
        directory : str, optional
            The directory where the results file will be saved. Default is "results".

        Notes
        -----
        .. note::

            The results are saved in a JSON file named "results_<identifier>.json" where <identifier>
            is a unique identifier for the calculation. The file contains the following keys:

            - "E_HOMO": The energy of the highest occupied molecular orbital (HOMO), rounded to 4 decimal places.
            - "E_LUMO": The energy of the lowest unoccupied molecular orbital (LUMO), rounded to 4 decimal places.
            - "HOMO": The list of HOMO values.
            - "LUMO": The list of LUMO values.
        """

        dictionary = {
            "E_HOMO": round(float(self.E_HOMO), 4),
            "E_LUMO": round(float(self.E_LUMO), 4),
            "HOMO": list(self.HOMO),
            "LUMO": list(self.LUMO),
        }

        filename = "results_" + self.identifier + ".json"
        filepath = os.path.join(directory, filename)

        with open(filepath, "w", encoding="utf-8") as f:
            json.dump(dictionary, f, indent=2)
        print("Results saved at" + filepath)

    def calc_MO(self, identifier, deviation=0):
        """
        Calculate the Molecular Orbital (MO) properties.

        Parameters
        ----------
        identifier : str
            The identifier for the MO to calculate. Should be either "HOMO" (Highest Occupied Molecular Orbital)
            or "LUMO" (Lowest Unoccupied Molecular Orbital).
        deviation : int, optional
            The deviation from the HOMO or LUMO level to calculate. Default is 0.

        Returns
        -------
        E_MO : float
            The energy of the specified MO.
        MO : numpy.ndarray
            The molecular orbital eigenvector.
        MO_occupation : numpy.ndarray
            The occupation of the molecular orbital.
        MO_type : str
            The type of the molecular orbital.
        MO_type_occupation : float
            The occupation of the molecular orbital type.
        """

        start = 0
        if identifier == "HOMO":
            start = -1 - deviation
        if identifier == "LUMO":
            start = deviation

        E_MO = self.eigv[self.num_electrons // 2 + start]
        MO = self.eigs[:, self.num_electrons // 2 + start]
        MO_occupation = MO.conj() * MO
        MO_type, MO_type_occupation = self._get_MO_type(MO_occupation)

        return E_MO, MO, MO_occupation, MO_type, MO_type_occupation

    # --------------------------------------------------------

    def _get_atom_distance_matrix(self):
        """Returns a matrix of distances between atoms in the molecule."""
        atom_distance_matrix = np.zeros((self.num_atoms, self.num_atoms))
        for i, j in combinations(range(self.num_atoms), r=2):
            value = np.linalg.norm(self.atom_coordinates[i] - self.atom_coordinates[j])
            atom_distance_matrix[i, j] = value
            atom_distance_matrix[j, i] = value
        return atom_distance_matrix

    def _get_orbital_distance_matrix(self):
        """Returns a matrix of distances between orbitals in the molecule."""
        orbital_distance_matrix = np.zeros((self.num_orbitals, self.num_orbitals))
        for i, j in combinations(range(self.num_orbitals), r=2):
            value = np.linalg.norm(
                self.orbital_coordinates[i] - self.orbital_coordinates[j]
            )
            orbital_distance_matrix[i, j] = value
            orbital_distance_matrix[j, i] = value
        return orbital_distance_matrix

    def _calc_center_of_mass(self):
        """Calculates the center of mass of the molecule."""
        molecule_mass = self.base1.mass + self.base2.mass
        center_of_mass = (
            self.base1.mass * self.base1.center_of_mass
            + self.base2.mass * self.base2.center_of_mass
        )
        center_of_mass /= molecule_mass
        return molecule_mass, center_of_mass

    def _calc_H(self):
        """Calculates the LCAO Hamiltonian matrix for the molecule."""
        H_int = np.zeros((self.base1.num_orbitals, self.base2.num_orbitals))
        for i, orbital1 in enumerate(self.base1.orbitals):
            for j, orbital2 in enumerate(self.base2.orbitals):
                value = calc_orbital_overlap(orbital1, orbital2, "interbase")
                H_int[i, j] = value
        H = np.block([[self.base1.H, H_int], [H_int.conj().T, self.base2.H]])
        return H

    # Molecular Orbitals
    def _get_MO_type(self, MO_occupation):
        """Returns the nature of the molecular orbitals: sigma, pi, n (non-bonding)."""

        identifiers = [
            identifier.split("_")[0] + "_" + identifier.split("_")[2]
            for identifier in self.orbital_identifiers
        ]
        s = ["C_s", "C_px", "C_py", "N_s", "O_s", "O_px", "O_py", "H_s"]
        s_mask = np.array([int(identifier in s) for identifier in identifiers])
        pi = ["C_pz", "N_pz", "O_pz"]
        pi_mask = np.array([int(identifier in pi) for identifier in identifiers])
        n = ["N_s", "N_px", "N_py"]
        n_mask = np.array([int(identifier in n) for identifier in identifiers])

        s_pop = sum(s_mask * MO_occupation)
        pi_pop = sum(pi_mask * MO_occupation)
        n_pop = sum(n_mask * MO_occupation)

        MO_type = ""
        if s_pop >= max(pi_pop, n_pop):
            MO_type = "sigma"

        if pi_pop >= max(s_pop, n_pop):
            MO_type = "pi"

        if n_pop >= max(s_pop, pi_pop):
            MO_type = "n"

        MO_type_occupation = [s_pop, pi_pop, n_pop]
        return MO_type, MO_type_occupation



# pylint: skip-file

import customtkinter as ctk

from .ham_frame import HamFrame
from .diss_frame import DissFrame
from .dynamics_frame import DynamicsFrame
from .user_interface_utils import change_state_all_widgets

# --------------------------------------------------


class OptionsTab(ctk.CTkTabview):
    def __init__(self, master, configs, **kwargs):
        """
        Notes:
            This tab is located in the options frame. This means that the options frame is its master.
            The tab itself is the master for three further frames.
        """

        # initialization of the ctk.CTkTabview class
        super().__init__(master)

        self.ham_tab = self.add("Hamiltonian")
        self.diss_tab = self.add("Dissipator")
        self.dynamics_tab = self.add("Dynamics")
        # self.set("Hamiltonian")

        self.ham_frame = HamFrame(self.ham_tab, configs, **kwargs)
        self.diss_frame = DissFrame(self.diss_tab, configs, **kwargs)
        self.dynamics_frame = DynamicsFrame(self.dynamics_tab, configs, **kwargs)


class OptionsFrame(ctk.CTkFrame):
    def __init__(self, master, configs, **kwargs):
        """
        Notes:
            This frame is located in the main window. This means that the main window is its master.
            The frame itself is the master for the tab options_tab.
            The frame uses the press_second_confirm(), enable_initial_frame() from the master.
        """

        # initialization of the ctk.CTkFrame class
        super().__init__(master)

        self.label = ctk.CTkLabel(
            self, text="Options", font=ctk.CTkFont(size=20, weight="bold")
        )
        self.label.grid(row=0, column=0, pady=10, padx=10, columnspan=2)

        self.options_tab = OptionsTab(self, configs, **kwargs)
        self.options_tab.grid(row=1, column=0, columnspan=2, pady=10, padx=10)

        self.back_button = ctk.CTkButton(
            self, text="Back", command=master.enable_initial_frame
        )
        self.back_button.grid(row=2, column=0, pady=20, padx=10)

        self.second_confirm_button = ctk.CTkButton(
            self, text="Confirm", command=master.press_second_confirm
        )
        self.second_confirm_button.grid(row=2, column=1, pady=10, padx=10)

    def change_state(self, state):
        """Changes the state of certain widgets (between 'normal' and 'disabled')."""
        change_state_all_widgets(self.options_tab.ham_frame, state=state)
        change_state_all_widgets(self.options_tab.diss_frame, state=state)
        change_state_all_widgets(self.options_tab.dynamics_frame, state=state)
        self.second_confirm_button.configure(state=state)
        self.back_button.configure(state=state)



# pylint: skip-file

import customtkinter as ctk

# --------------------------------------------------


class DissFrame(ctk.CTkFrame):
    def __init__(self, master, configs, **kwargs):
        """
        Notes:
            This frame is located in the options tab. This means that the options tab is its master.
        """

        # initialization of the ctk.CTkFrame class
        super().__init__(master)
        self.pack(fill="both", expand=True)
        self.configs = configs
        self.diss_kwargs_default = self.configs["diss_kwargs_default"]

        # widgets
        self.dephasing_label = ctk.CTkLabel(
            self, text="Dephasing", font=ctk.CTkFont(size=15, weight="bold")
        )
        self.dephasing_label.grid(row=0, column=0, pady=10, padx=10)

        self.diss_loc_deph_rate_label = ctk.CTkLabel(self, text="Local Dephasing Rate:")
        self.diss_loc_deph_rate_label.grid(row=1, column=0, padx=10, pady=10)

        self.diss_loc_deph_rate_entry = ctk.CTkEntry(self)
        self.diss_loc_deph_rate_entry.insert(
            0, str(self.diss_kwargs_default["loc_deph_rate"])
        )
        self.diss_loc_deph_rate_entry.grid(row=1, column=1, padx=10, pady=10)

        self.diss_glob_deph_rate_label = ctk.CTkLabel(
            self, text="Global Dephasing Rate:"
        )
        self.diss_glob_deph_rate_label.grid(row=2, column=0, padx=10, pady=10)

        self.diss_glob_deph_rate_entry = ctk.CTkEntry(self)
        self.diss_glob_deph_rate_entry.insert(
            0, str(self.diss_kwargs_default["glob_deph_rate"])
        )
        self.diss_glob_deph_rate_entry.grid(row=2, column=1, padx=10, pady=10)

        # -------------------------------

        self.relaxation_label = ctk.CTkLabel(
            self, text="Relaxation", font=ctk.CTkFont(size=15, weight="bold")
        )
        self.relaxation_label.grid(row=3, column=0, pady=10, padx=10)

        self.diss_relax_rate_label = ctk.CTkLabel(self, text="Relaxation Rate:")
        self.diss_relax_rate_label.grid(row=4, column=0, padx=10, pady=10)

        self.diss_relax_rate_entry = ctk.CTkEntry(self)
        # self.diss_relax_rate_entry.insert(
        #     0, str(self.diss_kwargs_default["relax_rate"])
        # )
        self.diss_relax_rate_entry.insert(0, 3.0)
        self.diss_relax_rate_entry.grid(row=4, column=1, padx=10, pady=10)

        # -----------------------------------

        self.thermalizing_label = ctk.CTkLabel(
            self, text="Thermalizing", font=ctk.CTkFont(size=15, weight="bold")
        )
        self.thermalizing_label.grid(row=5, column=0, pady=10, padx=10)

        self.diss_loc_therm_var = ctk.BooleanVar(
            value=self.diss_kwargs_default["loc_therm"]
        )
        self.diss_loc_therm_check = ctk.CTkCheckBox(
            self, text="Local", variable=self.diss_loc_therm_var
        )
        self.diss_loc_therm_check.grid(row=6, column=0, padx=10, pady=10)

        self.diss_glob_therm_var = ctk.BooleanVar(
            value=self.diss_kwargs_default["glob_therm"]
        )
        self.diss_glob_therm_check = ctk.CTkCheckBox(
            self, text="Global", variable=self.diss_glob_therm_var
        )
        self.diss_glob_therm_check.grid(row=6, column=1, padx=10, pady=10)

        self.diss_deph_rate_label = ctk.CTkLabel(self, text="Dephasing Rate:")
        self.diss_deph_rate_label.grid(row=7, column=0, padx=10, pady=10)

        self.diss_deph_rate_entry = ctk.CTkEntry(self)
        self.diss_deph_rate_entry.insert(0, str(self.diss_kwargs_default["deph_rate"]))
        self.diss_deph_rate_entry.grid(row=7, column=1, padx=10, pady=10)

        self.diss_cutoff_freq_label = ctk.CTkLabel(self, text="Cutoff Frequency:")
        self.diss_cutoff_freq_label.grid(row=8, column=0, padx=10, pady=10)

        self.diss_cutoff_freq_entry = ctk.CTkEntry(self)
        self.diss_cutoff_freq_entry.insert(
            0, str(self.diss_kwargs_default["cutoff_freq"])
        )
        self.diss_cutoff_freq_entry.grid(row=8, column=1, padx=10, pady=10)

        self.diss_reorg_energy_label = ctk.CTkLabel(self, text="Reorganization Energy:")
        self.diss_reorg_energy_label.grid(row=9, column=0, padx=10, pady=10)

        self.diss_reorg_energy_entry = ctk.CTkEntry(self)
        self.diss_reorg_energy_entry.insert(
            0, str(self.diss_kwargs_default["reorg_energy"])
        )
        self.diss_reorg_energy_entry.grid(row=9, column=1, padx=10, pady=10)

        self.diss_temperature_label = ctk.CTkLabel(self, text="Temperature (K):")
        self.diss_temperature_label.grid(row=10, column=0, padx=10, pady=10)

        self.diss_temperature_entry = ctk.CTkEntry(self)
        self.diss_temperature_entry.insert(
            0, str(self.diss_kwargs_default["temperature"])
        )
        self.diss_temperature_entry.grid(row=10, column=1, padx=10, pady=10)

        self.diss_spectral_density_label = ctk.CTkLabel(self, text="Spectral Density:")
        self.diss_spectral_density_label.grid(row=11, column=0, padx=10, pady=10)

        self.diss_spectral_density_combo = ctk.CTkComboBox(
            self, values=self.configs["SPECTRAL_DENSITIES"]
        )
        self.diss_spectral_density_combo.set(
            self.diss_kwargs_default["spectral_density"]
        )
        self.diss_spectral_density_combo.grid(row=11, column=1, padx=10, pady=10)

        self.diss_exponent_label = ctk.CTkLabel(self, text="Exponent:")
        self.diss_exponent_label.grid(row=12, column=0, padx=10, pady=10)

        self.diss_exponent_entry = ctk.CTkEntry(self)
        self.diss_exponent_entry.insert(0, str(self.diss_kwargs_default["exponent"]))
        self.diss_exponent_entry.grid(row=12, column=1, padx=10, pady=10)

    def set_uniform_relax(self):
        if self.diss_uniform_relaxation_var.get():
            self.diss_relax_rate_entry.configure(state="normal")
            self.diss_relax_rates_entry.configure(state="disabled")
        else:
            self.diss_relax_rate_entry.configure(state="disabled")
            self.diss_relax_rates_entry.configure(state="normal")

    def get_diss_kwargs(self):
        """Returns the values of widgets with get() method in dictionary format."""

        diss_kwargs = {
            "loc_deph_rate": float(self.diss_loc_deph_rate_entry.get()),
            "glob_deph_rate": float(self.diss_glob_deph_rate_entry.get()),
            "uniform_relaxation": True,
            "relax_rate": float(self.diss_relax_rate_entry.get()),
            "loc_therm": self.diss_loc_therm_var.get(),
            "glob_therm": self.diss_glob_therm_var.get(),
            "deph_rate": float(self.diss_deph_rate_entry.get()),
            "cutoff_freq": float(self.diss_cutoff_freq_entry.get()),
            "reorg_energy": float(self.diss_reorg_energy_entry.get()),
            "temperature": float(self.diss_temperature_entry.get()),
            "spectral_density": self.diss_spectral_density_combo.get(),
            "exponent": float(self.diss_exponent_entry.get()),
        }
        return diss_kwargs



"""This module provides functions to plot population and coherence for quantum DNA
models."""

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from qDNA.utils import calc_coherence, get_pop_fourier
from . import COLORS_PARTICLES

__all__ = [
    "plot_pops_heatmap",
    "plot_pop_fourier",
    "plot_pop",
    "plot_pops",
    "plot_coh",
    "plot_test_fourier",
]

# -------------------------------------------------------------------------------------


def plot_pops_heatmap(me_solver, vmax_list=[1, 1, 1], heatmap_type="seaborn"):
    """
    Plots a heatmap of populations for each particle in the system.

    Parameters
    ----------
    me_solver : object
        An instance of a solver that contains the time evolution data and Hamiltonian information.
    vmax_list : list of float, optional
        A list of maximum values for the color scale of each particle's heatmap. Default is [1, 1, 1].
    heatmap_type : str, optional
        The type of heatmap to plot. Options are "seaborn" or "matplotlib". Default is "seaborn".

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object containing the heatmaps.
    ax : numpy.ndarray of matplotlib.axes._subplots.AxesSubplot
        An array of axes objects for each particle's heatmap.

    Notes
    -----
    .. note::

        - The function uses seaborn or matplotlib to generate heatmaps.
        - The color maps used are "Blues", "Reds", and "Greens" for different particles.
    """

    num_particles = len(me_solver.tb_ham.particles)
    fig, ax = plt.subplots(
        1, num_particles, figsize=(6.4 * num_particles, 4.8), sharey=False
    )

    if len(me_solver.tb_ham.particles) == 1:
        ax = [ax]

    pop_dict = me_solver.get_pop()
    cmaps = ["Blues", "Reds", "Greys"]  # grey scale for exciton
    cmaps = ["Blues", "Reds", "Greens"]  # green scale for exciton
    for i, particle in enumerate(me_solver.tb_ham.particles):
        particle_pop = np.array(
            [value for key, value in pop_dict.items() if key.startswith(particle)]
        )

        # seaborn heatmap (looks prettier in my opinion)
        if heatmap_type == "seaborn":
            heatmap = sns.heatmap(
                particle_pop,
                xticklabels=[],
                yticklabels=[],
                cmap=cmaps[i],
                ax=ax[i],
                cbar=False,
                vmax=vmax_list[i],
            )
            heatmap.figure.colorbar(heatmap.collections[0], ax=ax[i])

        # matplotlib heatmap
        if heatmap_type == "matplotlib":
            im = ax[i].imshow(
                particle_pop, cmap=cmaps[i], aspect="auto", vmax=vmax_list[i]
            )
            fig.colorbar(im, ax=ax[i])

        ax[i].set_xticks([])
        ax[i].set_yticks([])

        y_len, x_len = particle_pop.shape
        ax[i].set_xlabel(f"Time [{me_solver.t_unit}]")
        ax[i].set_title(particle.capitalize())
        ax[i].set_xticks(
            np.linspace(0, x_len, 7), np.round(np.linspace(0, me_solver.t_end, 7), 0)
        )
        ax[i].set_yticks(
            np.arange(y_len) + 0.5, labels=me_solver.tb_ham.tb_sites_flattened
        )

    ax[0].set_ylabel("DNA Bases")
    return fig, ax


def plot_pop_fourier(ax, tb_ham, init_state, end_state, times, t_unit):
    """Plots population using the Fourier decomposition.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot on.
    tb_ham : TB_Ham
        The tight-binding Hamiltonian.
    init_state : str
        The initial state.
    end_state : str
        The end state.
    times : list of float
        The time points for plotting.
    t_unit : str
        The unit of time.
    """

    # calculation
    tb_ham.unit = "rad/" + t_unit
    amplitudes_dict, frequencies_dict, average_pop_dict = tb_ham.get_fourier(
        init_state, end_state, ["amplitude", "frequency", "average_pop"]
    )
    pop_dict = {}
    for particle in tb_ham.particles:
        amplitudes = amplitudes_dict[particle]
        frequencies = frequencies_dict[particle]
        average_pop = average_pop_dict[particle]
        pop_dict[particle] = [
            get_pop_fourier(t, average_pop, amplitudes, frequencies) for t in times
        ]

    # plotting
    for particle in tb_ham.particles:
        ax.plot(
            times, pop_dict[particle], label=particle, color=COLORS_PARTICLES[particle]
        )

    # plot settings
    ax.set_ylabel("Population")
    ax.set_xlabel("Time [" + t_unit + "]")
    ax.legend()
    return ax


def plot_pop(ax, tb_site, me_solver, add_legend=True):
    """Plots population of one base.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot on.
    tb_site : str
        The tight-binding site.
    me_solver : ME_Solver
        The master equation solver instance.
    add_legend : bool, optional
        Whether to add a legend, by default True.
    """

    # calculation
    description = me_solver.tb_ham.description
    if description == "2P":
        me_solver.get_pop()
    elif description == "1P":
        me_solver.get_result()

    # plotting
    particles = me_solver.tb_ham.particles
    for particle in particles:
        if description == "2P":
            ax.plot(
                me_solver.times,
                me_solver.pop[particle + "_" + tb_site],
                color=COLORS_PARTICLES[particle],
            )
        elif description == "1P":
            tb_basis = me_solver.tb_model.tb_basis
            tb_site_idx = tb_basis.index(tb_site)
            ax.plot(
                me_solver.times,
                [dm[tb_site_idx, tb_site_idx].real for dm in me_solver.result],
                color=COLORS_PARTICLES[particle],
            )

    # plot settings
    ax.set_ylim(0, 1.02)
    if add_legend:
        ax.set_ylabel("Population")
        ax.set_xlabel("Time [" + me_solver.t_unit + "]")
        ax.legend(particles)
    return ax


def plot_pops(me_solver):
    """Plots populations of all bases.

    Parameters
    ----------
    me_solver : ME_Solver
        The master equation solver instance.

    Returns
    -------
    tuple
        A tuple containing the figure and axes.
    """

    num_strands = me_solver.tb_model.num_strands
    num_sites_per_strand = me_solver.tb_model.num_sites_per_strand
    tb_basis = me_solver.tb_model.tb_basis

    # plotting
    fig, axes = plt.subplots(
        num_strands,
        num_sites_per_strand,
        figsize=(5 * num_sites_per_strand, 5 * num_strands),
        sharex=True,
        sharey=True,
    )
    axes = axes.flatten()
    for i, tb_site in enumerate(tb_basis):
        add_legend = i == 0
        axes[i] = plot_pop(axes[i], tb_site, me_solver, add_legend=add_legend)
    return fig, axes


def plot_coh(ax, me_solver):
    """Plots a measure of coherence as the sum of absolute values of the off- diagonal
    elements.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot on.
    me_solver : ME_Solver
        The master equation solver instance.
    """

    # calculation
    description = me_solver.tb_ham.description
    if description == "2P":
        me_solver.get_coh()
    elif description == "1P":
        me_solver.get_result()

    # plotting
    particles = me_solver.tb_ham.particles
    for particle in particles:
        if description == "2P":
            ax.plot(
                me_solver.times,
                me_solver.coh[particle],
                label=particle,
                color=COLORS_PARTICLES[particle],
            )
        elif description == "1P":
            ax.plot(
                me_solver.times,
                [calc_coherence(dm.full()) for dm in me_solver.result],
                label=particle,
                color=COLORS_PARTICLES[particle],
            )

    # plot settings
    ax.set_ylabel("Coherence")
    ax.set_xlabel("Time [" + me_solver.t_unit + "]")
    ax.legend()
    return ax


def plot_test_fourier(ax, tb_site, me_solver):
    """Plots a comparison of the result obtained by Fourier decomposition and the result
    obtained solving the Schrödinger equation with the qutip mesolver.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot on.
    tb_site : str
        The tight-binding site.
    me_solver : ME_Solver
        The master equation solver instance.
    """

    plot_pop_fourier(
        ax,
        me_solver.tb_ham,
        me_solver.init_state,
        tb_site,
        me_solver.times,
        me_solver.t_unit,
    )
    plot_pop(ax, tb_site, me_solver)
    return ax



# pylint: skip-file
import os

import customtkinter as ctk
from tkinter import filedialog, messagebox

from .. import DATA_DIR
from ..lcao import convert_pdb_to_xyz, calc_tb_params
from ..hamiltonian import wrap_save_tb_params

# --------------------------------------------------


class PDBFrame(ctk.CTkFrame):
    def __init__(self, master, configs):
        """
        Notes:
            This frame is located in the custom window. This means that the custom window is its master.
            The frame uses the commands save() and cancel() from the master

        Widgets with get() method:
            tb_params_entry, source_entry, particle_combobox, tb_model_combobox, unit_combobox, methylation_checkbox
        """

        # initialization of the ctk.CTkFrame class
        super().__init__(master)
        self.configs = configs

        # widgets

        # Upload button

        self.logo_label = ctk.CTkLabel(
            self, text="QuantumDNA", font=ctk.CTkFont(size=20, weight="bold")
        )
        self.logo_label.grid(row=0, column=0, pady=10, padx=10)

        self.upload_pdb_button = ctk.CTkButton(
            self, text="Upload PDB file", command=master.open_pdb_file
        )
        self.upload_pdb_button.grid(row=1, column=0, padx=10, pady=10)

        self.identifier_label = ctk.CTkLabel(self, text="Identifier:")
        self.identifier_label.grid(row=0, column=1, padx=10, pady=0)

        self.identifier_entry = ctk.CTkEntry(self)
        self.identifier_entry.grid(row=1, column=1, padx=10, pady=10)

        self.tb_model_label = ctk.CTkLabel(self, text="TB Model:")
        self.tb_model_label.grid(row=2, column=1, padx=10, pady=0)

        self.tb_model_combobox = ctk.CTkComboBox(self, values=self.configs["TB_MODELS"])
        self.tb_model_combobox.grid(row=3, column=1, padx=10, pady=10)
        self.tb_model_combobox.set("ELM")

        self.notes_label = ctk.CTkLabel(self, text="Notes:")
        self.notes_label.grid(row=4, column=1, padx=10, pady=0)

        self.notes_entry = ctk.CTkEntry(self)
        self.notes_entry.grid(row=5, column=1, padx=10, pady=10)

        self.text_label = ctk.CTkLabel(
            self,
            text="(Tip: Restart the session to apply and use \nthe calculated parameters in simulations.)",
        )
        self.text_label.grid(row=6, column=0, columnspan=2, pady=10, padx=10)

        self.save_button = ctk.CTkButton(self, text="Save", command=master.save)
        self.save_button.grid(row=7, column=0, padx=10, pady=10)

        self.cancel_button = ctk.CTkButton(self, text="Cancel", command=master.cancel)
        self.cancel_button.grid(row=7, column=1, padx=10, pady=10)


class PDBWindow(ctk.CTkToplevel):
    def __init__(self, master, configs):
        """
        Notes:
            This window is a toplevel window of the main window. This means that the main window is its master.
            The frame uses the commands save() and cancel() from the master
        """

        # initialization of the ctk.CTkToplevel class
        super().__init__(master)
        self.title("PDB Input")

        # add a instance of CustomFrame to the window
        self.custom_frame = PDBFrame(self, configs)
        self.custom_frame.grid(
            row=0, column=0, columnspan=2, padx=10, pady=10, sticky="nsew"
        )

    def open_pdb_file(self):
        pdb_file_path = filedialog.askopenfilename(
            title="Select PDB file", filetypes=[("Text files", "*.pdb")]
        )
        if not pdb_file_path:
            messagebox.showwarning("Warning", "No PDB file selected.")
            return

        self.pdb_file_path = pdb_file_path

    def get_custom_frame_params(self):
        """Makes all parameters of custom_frame available in this window."""
        self.identifier = self.custom_frame.identifier_entry.get()
        self.tb_model_name = self.custom_frame.tb_model_combobox.get()
        self.notes = self.custom_frame.notes_entry.get()

    def save(self):
        """Saves the tight-binding parameters and closes the window."""
        self.get_custom_frame_params()

        convert_pdb_to_xyz(self.pdb_file_path)
        # calculate tight-binding parameters
        directories = [self.pdb_file_path.split(".")[0]]
        HOMO_dict, LUMO_dict = calc_tb_params(directories, self.tb_model_name)

        wrap_save_tb_params(
            HOMO_dict,
            self.identifier,
            "hole",
            self.tb_model_name,
            unit="meV",
            notes=self.notes,
        )
        wrap_save_tb_params(
            LUMO_dict,
            self.identifier,
            "electron",
            self.tb_model_name,
            unit="meV",
            notes=self.notes,
        )
        self.destroy()

    def cancel(self):
        """Closes the window."""
        self.destroy()



"""This module is taken from quantum_HEOM (github.com/jwa7/quantum_HEOM), J.W. Abbott,
2022, DOI: 10.5281/zenodo.7230160.

It provides functions for calculating bath spectral densities and Lindblad rates.
"""

import numpy as np
import scipy.constants as c

# --------------------------- Bath Spectral Densities --------------------------------------


def debye_spectral_density(omega, cutoff_freq, reorg_energy):
    """Calculates the Debye spectral density.

    Parameters
    ----------
    omega : float
        Frequency.
    cutoff_freq : float
        Cutoff frequency.
    reorg_energy : float
        Reorganization energy.

    Returns
    -------
    float
        Debye spectral density.
    """
    if omega <= 0:
        return 0
    return 2 * reorg_energy * omega * cutoff_freq / (omega**2 + cutoff_freq**2)


def ohmic_spectral_density(omega, cutoff_freq, reorg_energy, exponent):
    """Calculates the Ohmic spectral density.

    Parameters
    ----------
    omega : float
        Frequency.
    cutoff_freq : float
        Cutoff frequency.
    reorg_energy : float
        Reorganization energy.
    exponent : float
        Specifies a sub- or superohmic bath.

    Returns
    -------
    float
        Ohmic spectral density.
    """
    if omega <= 0:
        return 0
    return (np.pi * reorg_energy * omega**exponent / cutoff_freq) * np.exp(
        -omega / cutoff_freq
    )


# ----------------------------- Lindblad Rates -------------------------------------------


def bose_einstein_distrib(omega, temperature):
    """Calculates the Bose-Einstein distribution.

    Parameters
    ----------
    omega : float
        Frequency.
    temperature : float
        Temperature.

    Returns
    -------
    float
        Bose-Einstein distribution.
    """
    assert (
        temperature != 0 and omega != 0
    ), "Temperature and frequency must be non-zero."
    return 1.0 / (np.exp(c.hbar * omega * 1e12 / (c.k * temperature)) - 1)


def rate_constant_redfield(
    omega,
    deph_rate,
    cutoff_freq,
    reorg_energy,
    temperature,
    spectral_density,
    exponent=None,
):
    """Calculates the Redfield rate constant.

    Parameters
    ----------
    omega : float
        Frequency.
    deph_rate : float or None
        Dephasing rate. If None, it will be calculated.
    cutoff_freq : float
        Cutoff frequency.
    reorg_energy : float
        Reorganization energy.
    temperature : float
        Temperature.
    spectral_density : str
        Spectral density type ('debye' or 'ohmic').
    exponent : float, optional
        Exponent for the Ohmic spectral density.

    Returns
    -------
    float
        Redfield rate constant.
    """
    if omega == 0:
        if deph_rate is None:
            deph_rate = dephasing_rate(cutoff_freq, reorg_energy, temperature)
        return deph_rate

    assert spectral_density in [
        "debye",
        "ohmic",
    ], "Spectral density must be 'debye' or 'ohmic'."

    spec_omega_ij, spec_omega_ji = 0, 0

    if spectral_density == "debye":
        spec_omega_ij = debye_spectral_density(omega, cutoff_freq, reorg_energy)
        spec_omega_ji = debye_spectral_density(-omega, cutoff_freq, reorg_energy)

    elif spectral_density == "ohmic":
        spec_omega_ij = ohmic_spectral_density(
            omega, cutoff_freq, reorg_energy, exponent
        )
        spec_omega_ji = ohmic_spectral_density(
            -omega, cutoff_freq, reorg_energy, exponent
        )

    n_omega_ij = bose_einstein_distrib(omega, temperature)
    n_omega_ji = bose_einstein_distrib(-omega, temperature)
    return 2 * ((spec_omega_ij * (1 + n_omega_ij)) + (spec_omega_ji * n_omega_ji))


def dephasing_rate(cutoff_freq, reorg_energy, temperature):
    """Calculates the dephasing rate in the limit of the Redfield rate equation as the
    frequency approaches zero.

    Parameters
    ----------
    cutoff_freq : float
        Cutoff frequency.
    reorg_energy : float
        Reorganization energy.
    temperature : float
        Temperature.

    Returns
    -------
    float
        Dephasing rate.
    """
    return (4 * reorg_energy * c.k * temperature) / (c.hbar * cutoff_freq * 1e12)



from .save_load import *
from .helpers import *

CONFIGS: dict = load_json("config", os.path.join(DATA_DIR, "raw"), load_metadata=False)
CONFIG: dict = {**DEFAULTS, **CONFIGS}
TB_MODELS_PROPS: dict = load_json("tb_models", os.path.join(DATA_DIR, "raw"))

DNA_BASES: list = CONFIGS["DNA_BASES"]
TB_MODELS: list = CONFIGS["TB_MODELS"]
SOURCES: list = CONFIGS["SOURCES"]
DESCRIPTIONS: list = CONFIGS["DESCRIPTIONS"]
PARTICLES: list = CONFIGS["PARTICLES"]
UNITS: list = CONFIGS["UNITS"]
T_UNITS: list = CONFIGS["T_UNITS"]
SPECTRAL_DENSITIES: list = CONFIGS["SPECTRAL_DENSITIES"]

from .check_input import *



"""This module provides functions for analyzing Hamiltonians in quantum systems,
including calculations of average population, amplitudes, frequencies, population using
Fourier series, and inverse participation ratio (IPR).

Notes
-----
- IPR: inverse participation ratio
- calc: calculate
- pop: population
- eigs: eigenstates
- val: value
- dims: dimensions
- init: initial
"""

from itertools import product

import numpy as np

__all__ = [
    "calc_average_pop",
    "calc_amplitudes",
    "calc_frequencies",
    "get_pop_fourier",
    "calc_ipr_hamiltonian",
]

# ------------------------------------------------


def calc_average_pop(eigs, init_state, end_state):
    """Calculates the time-averaged/ static population of the end state when initially
    in the initial state.

    Parameters
    ----------
    eigs : np.ndarray
        Eigenvector matrix.
    init_state : int
        Initial state in the basis.
    end_state : int
        End state in the basis.

    Returns
    -------
    float
        The static population of the end state.
    """
    average_pop = np.sum(eigs[end_state, :] ** 2 * eigs[init_state, :] ** 2)
    return average_pop


def calc_amplitudes(eigs, init_state, end_state):
    """Calculates the amplitudes for transitions between states.

    Parameters
    ----------
    eigs : np.ndarray
        Eigenvector matrix.
    init_state : int
        Initial state in the basis.
    end_state : int
        End state in the basis.

    Returns
    -------
    np.ndarray
        A list of amplitudes for transitions.
    """
    matrix_dimension = eigs.shape[0]
    amplitudes = [
        2
        * eigs[end_state, i]
        * eigs[init_state, i]
        * eigs[end_state, j]
        * eigs[init_state, j]
        for i, j in product(range(matrix_dimension), repeat=2)
        if i < j
    ]
    return np.real(np.array(amplitudes))


def calc_frequencies(eigv):
    """Calculates the frequencies corresponding to energy level differences.

    Parameters
    ----------
    eigv : np.ndarray
        Eigenvalue vector.

    Returns
    -------
    np.ndarray
        A list of frequencies.
    """
    matrix_dimension = len(eigv)
    frequencies = [
        abs(eigv[i] - eigv[j]).real
        for i, j in product(range(matrix_dimension), repeat=2)
        if i < j
    ]
    return np.array(frequencies)


def get_pop_fourier(t, average_pop, amplitudes, frequencies):
    """Calculates the population using Fourier series.

    Parameters
    ----------
    t : float
        Time variable.
    average_pop : float
        Average population.
    amplitudes : np.ndarray
        Amplitudes for transitions.
    frequencies : np.ndarray
        Frequencies corresponding to energy level differences.

    Returns
    -------
    float
        The population at time t.
    """
    population = average_pop
    for amplitude, frequency in zip(amplitudes, frequencies):
        population += amplitude * np.cos(frequency * t)
    return population


def calc_ipr_hamiltonian(eigs):
    r"""Calculates the inverse participation ratio (IPR) for each eigenstate of the
    Hamiltonian.

    Parameters
    ----------
    eigs : np.ndarray
        Eigenvector matrix.

    Returns
    -------
    list of float
        The IPR values for each eigenstate.

    Notes
    -----
    .. note::

        The IPR is a measure of the localization of an eigenstate. It is defined as the sum of the squared coefficients of the
        eigenvector. The IPR ranges from 1 to N, where N is the dimension of the Hilbert space. The IPR values can be used to
        distinguish between localized and delocalized eigenstates.

        .. math::
            \mathrm{IPR} = 1/\sum_{i=1}^{N} |c_i|^4.

        - Localized state: :math:`\mathrm{IPR} = 1/N`
        - Delocalized coherent state: :math:`\mathrm{IPR} = N`
    """

    dims = eigs.shape[0]
    ratios = []
    for idx in range(dims):
        ipr_val = 1 / np.sum(coeff**4 for coeff in eigs[:, idx])
        ratios.append(ipr_val)
    return ratios



"""
This module defines two classes for tight-binding models: Custom_TB_Model and TB_Model.

- Custom_TB_Model allows users to define a custom tight-binding model by specifying the model name, dimensions, basis states, and configurations.
- TB_Model provides a predefined tight-binding model based on the model name and dimensions, and retrieves the corresponding properties, configurations, and basis states.
"""

from ..tools import DEFAULTS
from .. import TB_MODELS_PROPS

from .tb_basis import get_tb_basis, get_eh_basis
from .tb_config import get_tb_config

__all__ = ["Custom_TB_Model", "TB_Model"]

# ------------------------------------------------------------------------------------------------------------


class Custom_TB_Model:
    """A class to define a custom tight-binding model.

    Parameters
    ----------
    tb_model_name : str
        The name of the tight-binding model.
    tb_dims : tuple
        A tuple specifying the dimensions of the model (number of strands, number of sites per strand).
    tb_basis : list of str
        A list of basis states for the tight-binding model.
    tb_config : list of tuple
        A list of configurations, where each configuration is a tuple of length three.

    Attributes
    ----------
    tb_model_name : str
        The name of the tight-binding model.
    tb_dims : tuple
        The dimensions of the model (number of strands, number of sites per strand).
    num_strands : int
        The number of strands in the model.
    num_sites_per_strand : int
        The number of sites per strand.
    num_sites : int
        The total number of sites in the model.
    tb_config : list of tuple
        The configurations of the model.
    tb_basis : list of str
        The basis states of the model.
    verbose : bool
        Verbose mode for printing debug information.
    """

    def __init__(self, tb_model_name, tb_dims, tb_basis, tb_config):
        # check inputs
        assert isinstance(tb_model_name, str), "tb_model_name must be of type str"
        assert (
            isinstance(tb_dims, tuple) and len(tb_dims) == 2
        ), "tb_dims must be of type tuple and of length two"
        assert all(
            isinstance(tb_dim, int) and tb_dim > 0 for tb_dim in tb_dims
        ), "elements of tb_dim must be of type int and > 0"
        assert isinstance(tb_basis, list), "tb_basis must be of type list"
        assert all(
            isinstance(tb_basis_element, str) for tb_basis_element in tb_basis
        ), "elements of tb_basis must be strings"
        assert isinstance(tb_config, list), "tb_config must be of type list"
        assert all(
            isinstance(tb_config_element, tuple) and len(tb_config_element) == 3
            for tb_config_element in tb_config
        ), "elements of tb_config must be of type tuple and of length three"

        self.verbose = DEFAULTS["verbose"]
        if self.verbose:
            print("Successfully checked all inputs of the Custom_TB_Model instance.")

        # set attributes
        self.tb_model_name = tb_model_name
        self.tb_dims = tb_dims
        self.num_strands, self.num_sites_per_strand = self.tb_dims
        self.num_sites = self.num_strands * self.num_sites_per_strand
        self.tb_config = tb_config
        self.tb_basis = tb_basis

        if self.verbose:
            print("Successfully initialized the Custom_TB_Model instance.")

    def __vars__(self) -> dict:
        """Returns the instance variables as a dictionary."""
        return vars(self)

    def __repr__(self) -> str:
        """Returns a string representation of the Custom_TB_Model instance."""
        return f"Custom_TB_Model({self.tb_model_name}, {self.tb_dims}, {self.tb_basis}, {self.tb_config})"

    def __eq__(self, other) -> bool:
        """Compares two Custom_TB_Model instances for equality."""
        return self.__repr__() == other.__repr__()


# ---------------------------------------------------------------------------------------------------------


class TB_Model:
    """A class to define a predefined tight-binding model.

    Parameters
    ----------
    tb_model_name : str
        The name of the predefined tight-binding model.
    tb_dims : tuple
        A tuple specifying the dimensions of the model (number of strands, number of sites per strand).

    Attributes
    ----------
    tb_model_name : str
        The name of the predefined tight-binding model.
    tb_dims : tuple
        The dimensions of the model (number of strands, number of sites per strand).
    num_strands : int
        The number of strands in the model.
    num_sites_per_strand : int
        The number of sites per strand.
    num_sites : int
        The total number of sites in the model.
    tb_model_props : dict
        The properties of the predefined tight-binding model.
    tb_config : list of tuple
        The configurations of the model.
    tb_basis : list of str
        The basis states of the model.
    eh_basis : list of str
        The electron-hole basis states of the model.
    verbose : bool
        Verbose mode for printing debug information.
    """

    def __init__(self, tb_model_name, tb_dims):
        # check inputs
        assert isinstance(tb_model_name, str), "tb_model_name must be of type str"
        assert (
            isinstance(tb_dims, tuple) and len(tb_dims) == 2
        ), "tb_dims must be of type tuple and of length two"
        assert all(
            isinstance(tb_dim, int) and tb_dim > 0 for tb_dim in tb_dims
        ), "elements of tb_dim must be of type int and > 0"
        self.verbose = DEFAULTS["verbose"]
        if self.verbose:
            print("Successfully checked all inputs of the TB_Model instance.")

        # set attributes
        self.tb_model_name = tb_model_name
        self.tb_dims = tb_dims
        self.num_strands, self.num_sites_per_strand = self.tb_dims
        self.num_sites = self.num_strands * self.num_sites_per_strand

        # take predefined model properties
        TB_MODELS = list(TB_MODELS_PROPS.keys())
        assert (
            self.tb_model_name in TB_MODELS
        ), f"tb_model_name must be a predefined model {TB_MODELS}"
        self.tb_model_props = TB_MODELS_PROPS[self.tb_model_name]
        assert (
            self.num_strands == self.tb_model_props["num_strands"]
        ), f"provided number of strands {self.num_strands} does not match the number of strands required for the predefined model {self.tb_model_props['num_strands']}"

        self.tb_config = get_tb_config(self.tb_model_name, self.tb_dims)
        self.tb_basis = get_tb_basis(self.tb_dims)
        self.eh_basis = get_eh_basis(self.tb_dims)

        if self.verbose:
            print("Successfully initialized the TB_Model instance.")

    def __vars__(self) -> dict:
        """Returns the instance variables as a dictionary."""
        return vars(self)

    def __repr__(self) -> str:
        """Returns a string representation of the TB_Model instance."""
        return f"TB_Model({self.tb_model_name}, {self.tb_dims})"

    def __eq__(self, other) -> bool:
        """Compares two TB_Model instances for equality."""
        return self.__repr__() == other.__repr__()



"""This module provides functions to calculate the estimated exciton lifetime for
quantum DNA models."""

import multiprocessing
import time
from functools import partial

import numpy as np
from tqdm import tqdm

from ..dynamics import get_me_solver
from ..tools import DEFAULTS, save_json

__all__ = ["calc_lifetime", "calc_lifetime_dict"]

# ---------------------------------------------------------------


def calc_lifetime(upper_strand, tb_model_name, **kwargs):
    """Calculates the exciton lifetime in femtoseconds (fs).

    Parameters
    ----------
    upper_strand : str
        The upper strand of DNA sequence.
    tb_model_name : str
        The name of the tight-binding model.
    kwargs : dict
        Additional keyword arguments for the master equation solver.

    Returns
    -------
    float or str
        The exciton lifetime in femtoseconds, or a message indicating no relaxation in the given time.

    Examples
    --------
    >>> calc_lifetime("GCG", "ELM", relax_rate=3, unit="rad/ps")
    775.5511022044088
    """
    start_time = time.time()
    me_solver = get_me_solver(upper_strand, tb_model_name, **kwargs)
    gs_pop = me_solver.get_groundstate_pop()["groundstate"]
    try:
        _, index = next((val, i) for i, val in enumerate(gs_pop) if val >= 1 - 1 / np.e)
        lifetime = me_solver.times[index]
        if me_solver.t_unit == "ps":
            lifetime *= 1000
        end_time = time.time()
        if DEFAULTS["verbose"]:
            print(f"Calculation time: {end_time-start_time}")
        return lifetime
    except StopIteration:
        return "no relaxation in the given time"


def calc_lifetime_dict(
    upper_strands, tb_model_name, filename, directory, num_cpu=None, **kwargs
):
    """Calculates the exciton lifetime for multiple upper strands using multiprocessing.

    Parameters
    ----------
    upper_strands : List[str]
        List of upper strands of DNA sequences.
    tb_model_name : str
        The name of the tight-binding model.
    filename : str
        The filename to save the lifetime dictionary.
    directory : str
        The directory where the lifetime dictionary is located.
    num_cpu : int, optional
        The number of CPU cores to use. Defaults to the total number of CPUs minus one.
    kwargs : dict
        Additional keyword arguments for the master equation solver.

    Returns
    -------
    Dict[str, float]
        Dictionary containing the exciton lifetime for each upper strand.
    """
    if not num_cpu:
        num_cpu = multiprocessing.cpu_count() - 1
    partial_calc_lifetime = partial(
        calc_lifetime, tb_model_name=tb_model_name, **kwargs
    )
    with multiprocessing.Pool(processes=num_cpu) as pool:
        lifetime_list = list(
            tqdm(
                pool.imap(partial_calc_lifetime, upper_strands),
                total=len(upper_strands),
                # file=sys.stdout,
            )
        )

    lifetime_dict = dict(zip(upper_strands, lifetime_list))
    save_json(
        lifetime_dict,
        kwargs,
        "lifetime_" + filename,
        directory,
    )
    return lifetime_dict



# pylint: skip-file

import customtkinter as ctk

# --------------------------------------------------


class InitialFrame(ctk.CTkFrame):
    def __init__(self, master, **kwargs):
        """
        Notes:
            This frame is located in the main/ menu window. This means that the main window is its master.
            The frame uses the commands press_first_confirm() from the master.

        Widgets with get() method:
            upper_strand_entry, tb_model_combo
        """

        # initialization of the ctk.CTkFrame class
        super().__init__(master)

        # widgets
        self.logo_label = ctk.CTkLabel(
            self, text="QuantumDNA", font=ctk.CTkFont(size=20, weight="bold")
        )
        self.logo_label.grid(row=0, column=0, pady=10, padx=10)

        self.grid_rowconfigure(1, weight=1)

        self.upper_strand_label = ctk.CTkLabel(
            self, text="Upper DNA Strand \n(5'-3' direction):"
        )
        self.upper_strand_label.grid(row=2, column=0, pady=0, padx=10)

        self.upper_strand_entry = ctk.CTkEntry(self)
        self.upper_strand_entry.grid(row=3, column=0, pady=10, padx=10)
        self.upper_strand_entry.insert(0, "G_C_G")

        self.lower_strand_label = ctk.CTkLabel(
            self, text="Lower DNA Strand \n(3'-5' direction):"
        )
        self.lower_strand_label.grid(row=4, column=0, pady=0, padx=10)

        self.lower_strand_entry = ctk.CTkEntry(self)
        self.lower_strand_entry.grid(row=5, column=0, pady=10, padx=10)
        self.lower_strand_entry.insert(0, "auto complete")

        self.tb_model_label = ctk.CTkLabel(self, text="TB Model:")
        self.tb_model_label.grid(row=6, column=0, pady=0, padx=10)

        self.tb_model_combo = ctk.CTkComboBox(self, values=master.configs["TB_MODELS"])
        self.tb_model_combo.grid(row=7, column=0, pady=10, padx=10)
        self.tb_model_combo.set("ELM")

        self.grid_rowconfigure(8, weight=1)

        self.first_confirm_button = ctk.CTkButton(
            self, text="Confirm", command=master.press_first_confirm
        )
        self.first_confirm_button.grid(row=9, column=0, pady=10, padx=10)

    def change_state(self, state):
        """Changes the state of certain widgets (between 'normal' and 'disabled')."""
        self.upper_strand_entry.configure(state=state)
        self.tb_model_combo.configure(state=state)
        self.first_confirm_button.configure(state=state)



import numpy as np
import scipy.constants as c

from .save_load import PARAMETRIZATION as PARAM

# ----------------------------------------


def calc_orbital_energy(orbital):
    """
    Calculate the energy of a given orbital.

    Parameters
    ----------
    orbital : Orbital
        An instance of the Orbital class containing information about the atom and orbital type.

    Returns
    -------
    float
        The energy of the specified orbital.
    """

    return PARAM["E_" + orbital.atom + orbital.orbital_type[0]]


def calc_orbital_overlap(
    orbital1, orbital2, connection
):  # pylint: disable=too-many-branches, too-many-statements
    """
    Calculate the overlap between two orbitals using the Harrison expression and the Slater-Koster two-center interactions.

    Parameters
    ----------
    orbital1 : Orbital
        The first orbital object containing its coordinates, atom type, orbital type, and atom identifier.
    orbital2 : Orbital
        The second orbital object containing its coordinates, atom type, orbital type, and atom identifier.
    connection : str
        The type of connection between the orbitals, either "interbase" or "intrabase".

    Returns
    -------
    float
        The calculated overlap between the two orbitals. Returns 0 if the distance between orbitals is zero,
        if the orbitals belong to the same atom in an intrabase connection, or if the distance exceeds the cutoff radius.

    Notes
    -----
    .. note::

        The directional cosines are used as projections on the coordinate axis rather than actual cosine values.
        The function includes a hydrogen correction factor obtained by optimization.
    """

    vector = np.array(orbital2.coordinates) - np.array(orbital1.coordinates)
    distance = np.linalg.norm(vector)

    if distance == 0:
        return 0

    vector_norm = vector / distance

    orbital_types = [orbital1.orbital_type, orbital2.orbital_type]

    # ----------------------------------------

    prefactor = None

    # interbase overlap, exponential decay term
    if connection == "interbase":
        if (orbital1.atom == "H") ^ (orbital2.atom == "H"):
            exponent = np.e ** (-2 / PARAM["d0H"] * (distance - PARAM["d0H"]))
            prefactor = (
                c.hbar**2 / (c.m_e * c.angstrom**2 * c.e * PARAM["d0H"] ** 2) * exponent
            )

        else:
            exponent = np.e ** (-2 / PARAM["d0"] * (distance - PARAM["d0"]))
            prefactor = (
                c.hbar**2 / (c.m_e * c.angstrom**2 * c.e * PARAM["d0"] ** 2) * exponent
            )

    # intrabase overlap, cutoff radius
    if connection == "intrabase":
        # exclude overlap between orbitals of the same atom
        if orbital1.atom_identifier == orbital2.atom_identifier:
            return 0

        # exclude overlap between orbitals that are too far apart
        if distance >= PARAM["cutoff_radius"]:
            return 0

        prefactor = c.hbar**2 / (c.m_e * c.angstrom**2 * c.e * distance**2)

    assert prefactor is not None, "Invalid connection type."

    # ----------------------------------------

    # hydrogen correction factor obtained by optimization
    correction = 1
    if orbital1.atom == "H":
        correction *= PARAM["b"]
    if orbital2.atom == "H":
        correction *= PARAM["b"]

    # ----------------------------------------

    overlap = None

    V_sssigma = PARAM["chi_sssigma"]
    V_spsigma = PARAM["chi_spsigma"]
    V_ppsigma = PARAM["chi_ppsigma"]
    V_pppi = PARAM["chi_pppi"]

    if orbital_types == ["s", "s"]:
        overlap = V_sssigma

    if orbital_types == ["s", "px"]:
        eta_1 = -vector_norm[0]
        overlap = V_spsigma * eta_1

    if orbital_types == ["px", "s"]:
        eta_1 = vector_norm[0]
        overlap = V_spsigma * eta_1

    if orbital_types == ["s", "py"]:
        eta_1 = -vector_norm[1]
        overlap = V_spsigma * eta_1

    if orbital_types == ["py", "s"]:
        eta_1 = vector_norm[1]
        overlap = V_spsigma * eta_1

    if orbital_types == ["s", "pz"]:
        eta_1 = -vector_norm[2]
        overlap = V_spsigma * eta_1

    if orbital_types == ["pz", "s"]:
        eta_1 = vector_norm[2]
        overlap = V_spsigma * eta_1

    # --------------------------------------

    if orbital_types in [["px", "py"], ["py", "px"]]:
        eta_1 = vector_norm[0]
        eta_2 = vector_norm[1]
        overlap = eta_1 * eta_2 * (V_ppsigma - V_pppi)

    if orbital_types in [["py", "pz"], ["pz", "py"]]:
        eta_1 = vector_norm[1]
        eta_2 = vector_norm[2]
        overlap = eta_1 * eta_2 * (V_ppsigma - V_pppi)

    if orbital_types in [["px", "pz"], ["pz", "px"]]:
        eta_1 = vector_norm[0]
        eta_2 = vector_norm[2]
        overlap = eta_1 * eta_2 * (V_ppsigma - V_pppi)

    if orbital_types == ["px", "px"]:
        eta_1 = vector_norm[0]
        overlap = eta_1**2 * V_ppsigma + (1 - eta_1**2) * V_pppi

    if orbital_types == ["py", "py"]:
        eta_1 = vector_norm[1]
        overlap = eta_1**2 * V_ppsigma + (1 - eta_1**2) * V_pppi

    if orbital_types == ["pz", "pz"]:
        eta_1 = vector_norm[2]
        overlap = eta_1**2 * V_ppsigma + (1 - eta_1**2) * V_pppi

    assert overlap is not None, "Invalid orbital types."

    return correction * prefactor * overlap



__version__ = "0.1.10"

import pathlib
import os

ROOT_DIR: str = str(pathlib.Path(__file__).absolute().parent.parent)
DATA_DIR: str = os.path.join(ROOT_DIR, "qDNA", "data")

from .tools import *
from .utils import *

from .lcao import *
from .dna_seq import *
from .model import *
from .hamiltonian import *
from .environment import *
from .dynamics import *
from .evaluation import *
from .visualization import *



import json
import os

import pandas as pd

from .. import DATA_DIR, DEFAULTS

__all__ = [
    "load_xyz",
    "PARAMETRIZATION",
    "find_xyz_files",
    "lcao_load_json",
    "convert_json_to_xyz",
    "convert_pdb_to_xyz",
]

# ----------------------------- JSON -----------------------------


def lcao_load_json(filepath):
    """Loads a JSON file."""
    try:
        with open(filepath, "r", encoding="utf-8") as f:
            data = json.load(f)
    except FileNotFoundError:
        message = f"File {filepath} does not exist."
        if DEFAULTS["verbose"]:
            print(message)
    return data


PARAMETRIZATION = lcao_load_json(
    os.path.join(
        DATA_DIR,
        "raw",
        "lcao_params",
        DEFAULTS["lcao_default"]["parametrization"] + ".json",
    )
)

# ----------------------------- XYZ -----------------------------


def find_xyz_files(directory):
    """
    Find all .xyz files in the given directory.
    Parameters
    ----------
    directory : str
        The path to the directory where .xyz files are to be searched.
    Returns
    -------
    list of str
        A list of filenames (without extensions) of all .xyz files found in the directory.
    Raises
    ------
    Exception
        If an error occurs while accessing the directory or reading its contents.
    """

    try:
        files = os.listdir(directory)
        xyz_files = [
            os.path.splitext(file)[0] for file in files if file.endswith(".xyz")
        ]
    except FileNotFoundError as e:
        print(f"File not found: {e}")
    return xyz_files


def load_xyz(filename, directory=os.getcwd()):
    """
    Load atomic coordinates from an XYZ file.

    Parameters
    ----------
    filename : str
        The name of the XYZ file (without the .xyz extension).
    directory : str, optional
        The directory where the XYZ file is located. Default is a subdirectory "geometries" within DATA_DIR.

    Returns
    -------
    tuple
        A tuple containing the identifier from the second line of the XYZ file and a
        DataFrame containing the atomic coordinates with columns ["Atom", "X", "Y", "Z"].

    Notes
    -----
    .. note::

        The XYZ file is expected to have the following format:
        - The first line contains the number of atoms (ignored).
        - The second line contains a comment or identifier.
        - Subsequent lines contain atomic coordinates in the format: Atom X Y Z.
    """

    filepath = os.path.join(directory, filename + ".xyz")

    with open(filepath, "r", encoding="utf-8") as file:
        # Skip the first line (number of atoms) and second line (comment)
        lines = file.readlines()

    xyz_identifier = lines[1].split()[0]
    xyz_data = pd.DataFrame(
        [line.split() for line in lines[2:]], columns=["Atom", "X", "Y", "Z"]
    )
    xyz_data[["X", "Y", "Z"]] = xyz_data[["X", "Y", "Z"]].astype(
        float
    )  # Convert coordinates to float
    return xyz_identifier, xyz_data


def convert_pdb_to_xyz(filepath_pdb):
    """
    Converts a PDB file to multiple XYZ files, one for each base.

    Parameters
    ----------
    filepath_pdb : str
        The path to the input PDB file.

    Notes
    -----
    .. note::

        Information about the atomic elements and coordinates is extracted from the PDB file.
        - The function creates a directory named after the input PDB file (without extension)
        to store the generated XYZ files.
        - Each base in the PDB file is written to a separate XYZ file.
        - If the chain identifier changes, the base numbering is adjusted to continue from
        the previous chain's last base number.
        - The function assumes that the base counter starts from one if the chain changes.
        - The function prints the directory where the XYZ files are created if the verbose
        mode is enabled in the DEFAULTS dictionary.

    """

    # Extract the filename without extension to create a folder
    filename = os.path.splitext(os.path.basename(filepath_pdb))[0]
    output_dir = os.path.join(os.path.dirname(filepath_pdb), filename)
    os.makedirs(output_dir, exist_ok=True)

    # Variables to hold atom data and track the current base
    elements = []
    coordinates = []
    current_base = None

    current_chain = ""
    current_base_number = 0

    # Read the PDB file
    lower_strand = False
    start_from_one = False
    num_bases_per_strand = 0

    with open(filepath_pdb, "r", encoding="utf-8") as file:
        for line in file:
            if line.startswith("ATOM") or line.startswith("HETATM"):
                # Parse element symbol and coordinates
                element = line[76:78].strip()  # Extract element symbol
                x = float(line[30:38].strip())  # Extract X coordinate
                y = float(line[38:46].strip())  # Extract Y coordinate
                z = float(line[46:54].strip())  # Extract Z coordinate

                # Parse base identifier
                base_identifier = line[17:20].strip()
                chain_identifier = line[21].strip()
                base_number = int(line[22:26].strip())

                chain_changes = current_chain != chain_identifier
                base_changes = current_base_number != base_number

                # checks if the base counter starts from one if the chain changes
                if chain_changes:
                    lower_strand = True
                    num_bases_per_strand = current_base_number
                    start_from_one = bool(base_number == 1)

                if base_changes and lower_strand and start_from_one:
                    base_number += num_bases_per_strand

                current_chain = chain_identifier
                current_base_number = base_number

                base_identifier = str(base_number).zfill(2) + base_identifier[1]

                # If base changes, write the previous base to a new .xyz file
                if current_base is not None and base_identifier != current_base:
                    write_xyz_file(output_dir, current_base, elements, coordinates)
                    elements = []
                    coordinates = []

                # Add current atom data
                elements.append(element)
                coordinates.append((x, y, z))
                current_base = base_identifier

        # Write the last base to a .xyz file
        if current_base is not None:
            write_xyz_file(output_dir, current_base, elements, coordinates)

        if DEFAULTS["verbose"]:
            print(f"XYZ files created in directory: {output_dir}")


def write_xyz_file(output_dir, base_identifier, elements, coordinates):
    """
    Write atomic coordinates to an XYZ file.

    Parameters
    ----------
    output_dir : str
        The directory where the XYZ file will be saved.
    base_identifier : str
        The base name for the XYZ file.
    elements : list of str
        A list of atomic element symbols.
    coordinates : list of tuple of float
        A list of tuples, each containing the x, y, and z coordinates of an atom.
    """

    filepath_xyz = os.path.join(output_dir, f"{base_identifier}.xyz")
    num_atoms = len(elements)

    # Prepare the content for the XYZ file
    xyz_content = f"{num_atoms}\n{base_identifier}\n"
    for element, (x, y, z) in zip(elements, coordinates):
        xyz_content += f"{element} {x:.4f} {y:.4f} {z:.4f}\n"

    # Write the content to the file
    with open(filepath_xyz, "w", encoding="utf-8") as file:
        file.write(xyz_content)


def convert_json_to_xyz(filename, directory):
    """
    Converts a JSON file to an XYZ file, e.g., from Pubchem https://pubchem.ncbi.nlm.nih.gov/.

    Parameters
    ----------
    filename : str
        The name of the JSON file (without extension) to be converted.
    directory : str
        The directory where the JSON file is located and where the XYZ file will be saved.

    Raises
    ------
    FileNotFoundError
        If the JSON file does not exist in the specified directory.
    KeyError
        If the expected keys are not found in the JSON file.
    ValueError
        If the JSON file contains invalid data.

    Notes
    -----
    .. note::
        The JSON file is expected to follow a specific structure with atomic data under
        "PC_Compounds" -> "atoms" and coordinates under "coords" -> "conformers".
        The atomic symbols are mapped from atomic numbers using a predefined dictionary.
        The JSON file is removed after conversion.

    Examples
    --------
    >>> convert_json_to_xyz("molecule", "/path/to/directory")
    File successfully converted and saved at /path/to/directory/molecule.xyz
    """

    # Input file path
    filepath_json = os.path.join(directory, filename + ".json")

    # Read the JSON file
    json_file = lcao_load_json(filepath_json)

    num_atoms = len(json_file["PC_Compounds"][0]["atoms"]["aid"])
    elements = json_file["PC_Compounds"][0]["atoms"]["element"]
    coordinates = json_file["PC_Compounds"][0]["coords"][0]["conformers"][0]
    x_coords = coordinates["x"]
    y_coords = coordinates["y"]
    z_coords = coordinates["z"]
    atomic_symbols = {1: "H", 6: "C", 7: "N", 8: "O"}

    # Remove the JSON file
    os.remove(filepath_json)

    # Prepare the content for the XYZ file
    xyz_content = f"{num_atoms}\n{filename}\n"  # Header
    for i in range(num_atoms):
        element_symbol = atomic_symbols.get(
            elements[i], "X"
        )  # Default to 'X' if not found
        x = x_coords[i]
        y = y_coords[i]
        z = z_coords[i]
        xyz_content += f"{element_symbol} {x:.4f} {y:.4f} {z:.4f}\n"

    # Output file path
    filepath_xyz = os.path.join(directory, filename + ".xyz")

    # Write the XYZ file
    with open(filepath_xyz, "w", encoding="utf-8") as file:
        file.write(xyz_content)

    if DEFAULTS["verbose"]:
        print(f"File successfully converted and saved at {filepath_xyz}")



# pylint: skip-file

import json
import glob
import os
import multiprocessing
import time
from tkinter import filedialog, messagebox

import customtkinter as ctk
import pandas as pd

from .. import DATA_DIR
from ..tools import load_json
from ..evaluation import (
    calc_dipole_dict,
    calc_dipole_moment_dict,
    calc_exciton_transfer_dict,
    calc_lifetime_dict,
)
from .scrollable_console_frame import ScrollableConsoleFrame

# --------------------------------------------------


def format_time(seconds):
    if seconds >= 3600:
        hours = seconds // 3600
        minutes = (seconds % 3600) // 60
        seconds = seconds % 60
        return f"{hours}h {minutes}m {seconds}s" if minutes or seconds else f"{hours}h"
    elif seconds >= 60:
        minutes = seconds // 60
        seconds = seconds % 60
        return f"{minutes}m {seconds}s" if seconds else f"{minutes}m"
    else:
        return f"{seconds}s"


class FastaWindow(ctk.CTkToplevel):
    def __init__(self, master):
        super().__init__(master)

        self.title("Fasta Input")
        self.setup_ui()

        self.upper_strands = []
        self.identifiers = []

        self.fasta_data = []
        self.kwargs = {}
        self.tb_model_name = "ELM"
        self.filename = "delete"
        self.num_cpu = multiprocessing.cpu_count() - 1
        self.directory = os.path.join(DATA_DIR, "gui")

    def setup_ui(self):
        self.logo_label = ctk.CTkLabel(
            self, text="QuantumDNA", font=ctk.CTkFont(size=20, weight="bold")
        )
        self.logo_label.grid(row=0, column=0, pady=10, padx=10)

        # Upload buttons
        self.upload_fasta_button = ctk.CTkButton(
            self, text="Upload FASTA file", command=self.open_fasta_file
        )
        self.upload_fasta_button.grid(row=1, column=0, padx=10, pady=10, sticky="w")

        self.upload_kwargs_button = ctk.CTkButton(
            self, text="Upload Options file", command=self.open_kwargs_file
        )
        self.upload_kwargs_button.grid(row=2, column=0, padx=10, pady=10, sticky="w")

        # Calculation options
        self.lifetime_var = ctk.BooleanVar(value=True)
        self.lifetime_checkbox = ctk.CTkCheckBox(
            self,
            text="Calculate Exciton Lifetime",
            state="disabled",
            variable=self.lifetime_var,
        )
        self.lifetime_checkbox.grid(row=0, column=1, padx=10, pady=10, sticky="w")

        self.dipole_var = ctk.BooleanVar(value=True)
        self.dipole_checkbox = ctk.CTkCheckBox(
            self, text="Calculate Charge Separation", variable=self.dipole_var
        )
        self.dipole_checkbox.grid(row=1, column=1, padx=10, pady=10, sticky="w")

        self.dipole_moment_var = ctk.BooleanVar(value=True)
        self.dipole_moment_checkbox = ctk.CTkCheckBox(
            self, text="Calculate Dipole Moment", variable=self.dipole_moment_var
        )
        self.dipole_moment_checkbox.grid(row=2, column=1, padx=10, pady=10, sticky="w")

        self.exciton_transfer_var = ctk.BooleanVar(value=True)
        self.exciton_transfer_checkbox = ctk.CTkCheckBox(
            self,
            text="Calculate Exciton Population",
            variable=self.exciton_transfer_var,
        )
        self.exciton_transfer_checkbox.grid(
            row=3, column=1, padx=10, pady=10, sticky="w"
        )

        self.estimate_comp_time_button = ctk.CTkButton(
            self, text="Estimate Computation Time", command=self.estimate_comp_time
        )
        self.estimate_comp_time_button.grid(
            row=4, column=0, columnspan=2, padx=10, pady=10, sticky="ew"
        )

        # Submit button
        self.submit_button = ctk.CTkButton(
            self, text="Submit", command=self.process_files
        )
        self.submit_button.grid(
            row=5, column=0, columnspan=2, padx=10, pady=10, sticky="ew"
        )

        self.scrollable_console_frame = ScrollableConsoleFrame(self)
        self.scrollable_console_frame.grid(
            row=6, column=0, columnspan=2, padx=10, pady=10, sticky="nsew"
        )

    def open_fasta_file(self):
        fasta_file_path = filedialog.askopenfilename(
            title="Select FASTA file", filetypes=[("Text files", "*.fasta")]
        )
        if not fasta_file_path:
            messagebox.showwarning("Warning", "No FASTA file selected.")
            return

        try:
            with open(fasta_file_path, "r") as file:
                self.fasta_data = file.readlines()
                self.parse_fasta_data()
            if not self.fasta_data:
                raise ValueError("FASTA file is empty.")
        except Exception as e:
            messagebox.showerror("Error", f"Failed to open FASTA file: {e}")
            self.fasta_data = []

    def open_kwargs_file(self):
        kwargs_file_path = filedialog.askopenfilename(
            title="Select Options file", filetypes=[("JSON files", "*.json")]
        )
        if not kwargs_file_path:
            messagebox.showwarning("Warning", "No Options file selected.")
            return

        try:
            with open(kwargs_file_path, "r") as json_file:
                self.kwargs = json.load(json_file)
            if not self.kwargs:
                raise ValueError("Options file is empty.")
        except Exception as e:
            messagebox.showerror("Error", f"Failed to open Options file: {e}")
            self.kwargs = {}

    def process_files(self):
        if not self.fasta_data:
            messagebox.showwarning("Warning", "Please provide a FASTA .fasta file")
            return

        fasta_data = {
            "Identifier": self.identifiers,
            "Sequence (5'-3')": self.upper_strands,
        }

        try:
            self.tb_model_name = self.kwargs["tb_model_name"]
            del self.kwargs["tb_model_name"]
        except:
            pass

        # start calculations
        start_time = time.time()

        # Lifetime calculation
        if self.lifetime_var.get():
            lifetime_dict = calc_lifetime_dict(
                self.upper_strands,
                self.tb_model_name,
                self.filename,
                self.directory,
                **self.kwargs,
            )
            fasta_data["Exciton Lifetime (fs)"] = list(lifetime_dict.values())
        end_time_lifetime = time.time()

        # Charge separation calculation
        if self.dipole_var.get():
            dipole_dict = calc_dipole_dict(
                self.tb_model_name, self.filename, self.directory
            )
            fasta_data["Charge Separation (A)"] = list(dipole_dict.values())

        # Dipole moment calculation
        if self.dipole_moment_var.get():
            dipole_moment_dict = calc_dipole_moment_dict(
                self.tb_model_name, self.filename, self.directory
            )
            fasta_data["Dipole Moment (D)"] = list(dipole_moment_dict.values())

        # Exciton transfer calculation
        if self.exciton_transfer_var.get():
            exciton_transfer_dict = calc_exciton_transfer_dict(
                self.tb_model_name, self.filename, self.directory
            )
            fasta_data["Exciton Transfer (upper strand)"] = [
                val[0]["exciton"] for val in exciton_transfer_dict.values()
            ]
            fasta_data["Exciton Transfer (lower strand)"] = [
                val[1]["exciton"] for val in exciton_transfer_dict.values()
            ]
        end_time = time.time()
        # end calculations

        # print("-------------------")
        # print(fasta_data)
        # print("-------------------")
        df_fasta = pd.DataFrame(fasta_data)
        self.kwargs["Computation Time (s)"] = end_time - start_time
        self.kwargs["Computation Time Lifetime (s)"] = end_time_lifetime - start_time
        metadata_df = pd.DataFrame(
            list(self.kwargs.items()), columns=["Parameter", "Value"]
        )
        # save as excel file
        self.save_to_excel(df_fasta, metadata_df)

        # delete files created during calculations
        json_files = glob.glob(os.path.join(self.directory, "*.json"))
        for file in json_files:
            os.remove(file)
            # print(f"Deleted: {file}")

    def parse_fasta_data(self):
        for line in self.fasta_data:
            line = line.strip()
            if line.startswith(">"):
                self.identifiers.append(line)
            else:
                self.upper_strands.append(line)

    def save_to_excel(self, df_fasta, metadata_df):
        save_path = filedialog.asksaveasfilename(
            defaultextension=".xlsx", filetypes=[("Excel files", "*.xlsx")]
        )
        if not save_path:
            messagebox.showwarning("Warning", "No save path specified.")
            return

        try:
            with pd.ExcelWriter(save_path) as writer:
                df_fasta.to_excel(writer, sheet_name="Data", index=False)
                metadata_df.to_excel(writer, sheet_name="Metadata", index=False)

            messagebox.showinfo("Success", f"File saved to {save_path}")
        except Exception as e:
            messagebox.showerror("Error", f"Failed to save file: {e}")

    def estimate_comp_time(self):
        comp_time_dict = load_json("comp_time", os.path.join(DATA_DIR, "raw"))
        comp_time = 0
        for upper_strand in self.upper_strands:
            comp_time += comp_time_dict[self.tb_model_name][len(upper_strand) - 2]
        comp_time /= min(self.num_cpu, len(self.upper_strands))
        print(
            "Estimated Computation Time "
            + format_time(comp_time)
            + f"\nUsing {self.num_cpu} kernels"
        )
